#!/usr/bin/env python

# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module-level script dispatcher for BASEmesh.

This script acts as a shortcut to access other scripts and shared
utilities available in BASEmesh.

Enter `--help` for a list of all available modules, along with a brief
description.

"""

import argparse
import subprocess
import sys

from basemesh import __version__ as version


if __name__ == '__main__':
    print('\n'
          '---------------------------------------\n'
          f'BASEmesh - Version {version}\n'
          '---------------------------------------\n')

    # Set up argument parser
    parser = argparse.ArgumentParser(
        'BASEmesh', description='Module-level script dispatcher', epilog='\n')
    # Define submodules
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--1D', dest='target', action='store_const', const='basechange',
        help='BASEchange 1D channel generator')

    # Consume remaining arguments
    #
    # NOTE: This will not catch any exact duplicates at the very beginning of
    # the command, e.g. "python -m basemesh --1D --1D" would both be "eaten" by
    # by this script.
    # However, "python -m basemesh --1D -something else --1D" will be forwarded
    # as expected.
    parser.add_argument('command', nargs=argparse.REMAINDER)

    args = parser.parse_args()

    # Run the associated script
    subprocess.call(
        (sys.executable, '-m', f'basemesh.{args.target}', *args.command))



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Python implementation of the about widget GUI."""

import os
from typing import Optional

from PyQt5.QtGui import QPixmap, QResizeEvent
from PyQt5.QtWidgets import QLabel, QPushButton, QTextBrowser
from qgis.gui import QgisInterface  # type: ignore

from .. import utils
from .base import DialogBase

try:
    import markdown2
except ModuleNotFoundError:
    import sys
    wheel_dir = os.path.join(utils.get_plugin_dir(), 'packages',
                             'markdown2-2.3.8-py2.py3-none-any.whl')
    sys.path.append(wheel_dir)
    import markdown2  # type: ignore


class AboutDialog(DialogBase):
    """GUI implementation of the elevation meshing utility."""

    # NOTE: Qt uses camelCase for its attributes - too bad!
    # pylint: disable=invalid-name

    ui_file = 'ui_about.ui'

    def __init__(self, iface: QgisInterface) -> None:
        """Connect signal callbacks and populate widgets.

        Parameters
        ----------
        iface : QgisInterface
            The QGIS interface to create the dialog for

        """
        super().__init__(iface)  # Initialize DialogBase

        # NOTE: The following are attribute type hints used to inform supported
        # IDEs of the attribute types defined in the Qt UI file. This code is
        # executed during runtime, but does not do anything as far as the
        # Python interpreter is concerned.
        self.banner: QLabel
        self.textBrowser: QTextBrowser
        self.linkRepo: QLabel
        self.linkWebsite: QLabel
        self.closeBtn: QPushButton

        # Load about text
        about_file = os.path.join(utils.get_plugin_dir(), 'plugin', 'gui',
                                  'templates', 'about_basemesh.md')
        try:
            with open(about_file) as about_markdown:
                html_str: str = markdown2.markdown(about_markdown.read())
        except FileNotFoundError:
            self.textBrowser.setText(
                f'Unable to generate help HTML, file not found:\n{about_file}')
        else:
            self.textBrowser.setHtml(html_str)

        # Apply size constraints
        self.pixmap = QPixmap(
            ':/plugins/basemesh/plugin/icons/basemesh_logo_with_text.png')
        self.resizeEvent()

    def resizeEvent(self, _: Optional[QResizeEvent] = None) -> None:
        """Event hook for the dialog being resized by the user."""
        # NOTE: Qt Designer does not support any form of "keep aspect ratio"
        # constraint, so this is set up in code instead.
        scaled = self.pixmap.scaledToHeight(self.banner.height())
        self.banner.setPixmap(scaled)



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Helper mesh factories.

These factories are still abstract, but provide simplified endpoints
for common factory designs.
"""

import math
from typing import List, Tuple
from ..core.geometry import Node, Segment
from ..types import Point2D
from .core import BasicFactory


class CircularMeshFactory(BasicFactory):
    """A factory for meshes with circular domains.

    This is a subclass of BasicFactory that also generates a circular
    mesh domain as part of its initialiser.

    The user can either subclass this factory to define their elevation
    function or use the elevation_func method to set the function for
    an initialised factory.
    """

    def __init__(self, radius: float, segments: int = 20,
                 midpoint_offset: Point2D = (0.0, 0.0)) -> None:
        """Initialise a mesh factory with a circular domain.

        Note that the domain cannot actually be perfectly circular due
        to the triangulation. Use the segments argument to specify the
        number of edges of the polygon used to approximate a circle to
        the accuracy your model requires.

        Parameters
        ----------
        radius : float
            The radius of the circle
        segments : int, optional
            The number of line segments used to approximate the circle,
            by default 20
        midpoint_offset : Tuple[float, float], optional
            X and Y offsets for the circle's midpoint; use this to
            position the circle, by default (0.0, 0.0)

        Raises
        ------
        ValueError
            Raised if the given radius is less than or equal to zero
        """
        super().__init__()
        if radius <= 0.0:
            raise ValueError('Domain radius must be greater than zero')
        # Create a polygon approximating the given input circle
        vertices = self._polygonise_circle(radius, segments, *midpoint_offset)
        # Convert vertices to nodes
        nodes = [Node(p) for p in vertices]
        # Add the first node a second time to simplify the loop
        nodes.append(nodes[0])
        # Add mesh boundary break lines
        self.segments = []
        for index, node in enumerate(nodes[:-1]):
            self.segments.append(Segment(node, nodes[index+1]))

    @staticmethod
    def _polygonise_circle(radius: float, sides: int = 12,
                           offset_x: float = 0.0,
                           offset_y: float = 0.0) -> List[Point2D]:
        """Return a regular polygon approximating the given circle.

        The polygon's vertices will lie on the circle, its edges will
        therefore intersect the circle.

        Parameters
        ----------
        radius : float
            The radius of the circle
        sides : int, optional
            The number of sides of the polygon, by default 12
        offset_x : float, optional
            Shift the midpoint in the X direction, by default 0.0
        offset_y : float, optional
            Shift the midpoint in the Y direction, by default 0.0

        Returns
        -------
        List[Tuple[float, float]]
            The polygon vertices
        """
        vertices = []
        for index in range(0, sides):
            # Calculate the angle of the given point
            angle = index/sides * 2 * math.pi
            # Get the x and y coordinates for the given angle
            pos_x = (math.sin(angle) * radius) + offset_x
            pos_y = (math.cos(angle) * radius) + offset_y
            # Add the vertex
            vertices.append((pos_x, pos_y))
        return vertices


class RectangularMeshFactory(BasicFactory):
    """A factory for meshes with rectangular domains.

    This is a subclass of BasicFactory that also generates a
    rectangular mesh domain as part of its initialiser.

    The user can either subclass this factory to define their elevation
    function or use the elevation_func method to set the function for
    an initialised factory.
    """

    def __init__(self, width: float, height: float,
                 midpoint_offset: Tuple[float, float] = (0.0, 0.0)) -> None:
        """Initialise the mesh factory.

        Parameters
        ----------
        width : float
            The width of the rectangle (X axis)
        height : float
            The height of the rectange (Y axis)
        midpoint_offset : Tuple[float, float], optional
             X and Y offsets for the rectangle's midpoint; use this to
            position the rectangle, by default (0.0, 0.0)

        Raises
        ------
        ValueError
            Raised if either edge is shorter than or equal to zero
        """
        super().__init__()
        if width <= 0.0 or height <= 0.0:
            raise ValueError('Domain edge length must be greater than zero')
        # Calculate corner vertices
        offset_x, offset_y = midpoint_offset
        min_x = offset_x - width/2
        max_x = offset_x + width/2
        min_y = offset_y - height/2
        max_y = offset_y + height/2
        vertices = [Node((min_x, min_y)), Node((max_x, min_y)),
                    Node((max_x, max_y)), Node((min_x, max_y))]
        # Add the first node a second time to simplify the loop
        vertices.append(vertices[0])
        # Add mesh boundary break lines
        self.segments = []
        for index, node in enumerate(vertices[:-1]):
            self.segments.append(Segment(node, vertices[index+1]))



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Processing algorithms contributed by BASEmesh."""

import os
from typing import Any
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtGui import QIcon
from qgis.core import QgsProcessingProvider  # type: ignore
from ..utils import get_plugin_dir
from . import converters, matid

__all__ = [
    'BasemeshProcessingProvider'
]


class BasemeshProcessingProvider(QgsProcessingProvider):
    """Processing algorithm provider for BASEmesh."""

    def loadAlgorithms(self) -> None:
        """Load all algorithms provided by the provider."""
        self.addAlgorithm(converters.LegacyLineTo3DLine())
        self.addAlgorithm(converters.LegacyPointTo3DPoint())
        self.addAlgorithm(matid.AssignMaterialID())
        self.addAlgorithm(matid.ExtractMaterialID())

    @staticmethod
    def id() -> str:
        """Return the unique processing provider ID."""
        return 'basemesh'

    @staticmethod
    def name() -> str:
        """Return the user-friendly name of the processing provider."""
        return 'BASEmesh'

    @staticmethod
    def icon() -> Any:
        """Return the QIcon to use for the processing provider."""
        return QIcon(os.path.join(
            get_plugin_dir(), 'plugin', 'icons', 'basemesh_icon.svg'))



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Shared base classes for the BASEmesh GUI."""

import datetime
import os
import time
import traceback
from typing import Any, Callable, Deque, Optional, Tuple
from PyQt5.QtCore import QThread
from PyQt5.QtWidgets import (QComboBox, QDialog, QFileDialog, QLineEdit,
                             QMessageBox, QPlainTextEdit, QPushButton,
                             QTextEdit)
from PyQt5.uic import loadUi  # type: ignore
from qgis.core import (QgsApplication, QgsMapLayer,  # type: ignore
                       QgsMapLayerType, QgsWkbTypes)
from qgis.gui import QgisInterface  # type: ignore
from ..utils import (add_layers, get_plugin_dir, get_project_dir,
                     get_project_title, layer_is_3d)
from .worker import Worker

try:
    import markdown2
except ModuleNotFoundError:
    import sys
    wheel_dir = os.path.join(get_plugin_dir(), 'packages',
                             'markdown2-2.3.8-py2.py3-none-any.whl')
    sys.path.append(wheel_dir)
    import markdown2  # type: ignore

# Type aliases
_FeedbackSignal = Tuple[Optional[float], Optional[str], bool]


class DialogBase(QDialog):
    """Base class for standalone windows in BASEmesh.

    Any subclasses must overwrite the ui_file name with the name of
    the .ui file to use for their initialisation.
    """

    # NOTE: While it would be nice to turn this into an ABC to ensure certain
    # class attributes are overridden before initialization, this is not
    # possible due to the way Qt wrapper classes are instantiated.
    # I am sure there are workarounds for this, but they would probably cause
    # more confusion and visual clutter than the use-case justifies.

    ui_file: str  # The UI resource to use
    help_file: str  # Markdown file containing help text (optional)

    def __init__(self, iface: QgisInterface) -> None:
        """Set the interface attribute and populate the dialog.

        Parameters
        ----------
        iface : QgisInterface
            The QGIS interface object to load itno

        Raises
        ------
        AttributeError
            Raised if the ui_file attribute has not been overwritten
            with a custom path
        """
        super().__init__()  # Initialize QDialog
        self.iface = iface
        # Load UI resource file
        try:
            src_path = os.path.join(
                get_plugin_dir(), 'plugin', 'gui', 'templates', self.ui_file)
        except AttributeError as err:
            raise AttributeError(f'{self.__class__.__name__} does not provide '
                                 'an ui_file attribute') from err
        loadUi(src_path, self)

    def add_mesh_layers(self, combo_box: QComboBox, *args: QComboBox) -> None:
        """Populate the combo box with mesh layers.

        Use QComboBox.currentData() to retrieve the currently selected
        layer.
        The combo box will be cleared as part of this call.

        Parameters
        ----------
        combo_box : QComboBox
            The combo box to add the mesh layers to
        """
        icon = QgsApplication.getThemeIcon('/mIconMeshLayer.svg')
        add_layers(self.iface.mapCanvas(), combo_box, *args, icon=icon,
                   filter_=lambda x: x.type()  # type: ignore
                   == QgsMapLayerType.MeshLayer)

    def add_point_layers(self, combo_box: QComboBox, *args: QComboBox,
                         has_z: Optional[bool] = None) -> None:
        """Populate the combo box with point vector layers.

        Use QComboBox.currentData() to retrieve the currently selected
        layer.
        The combo box will be cleared as part of this call.

        Parameters
        ----------
        combo_box : QComboBox
            The combo box to add the vector layers to
        """

        def filter_(layer: QgsMapLayer) -> bool:
            if (layer.type() != QgsMapLayerType.VectorLayer
                    or layer.geometryType() != QgsWkbTypes.PointGeometry):
                return False
            return has_z is None or has_z == layer_is_3d(layer)

        icon = QgsApplication.getThemeIcon('/mIconPointLayer.svg')
        add_layers(self.iface.mapCanvas(), combo_box, *args, icon=icon,
                   filter_=filter_)

    def add_polygon_layers(self, combo_box: QComboBox, *args: QComboBox,
                           has_z: Optional[bool] = None) -> None:
        """Populate the combo box with polygon vector layers.

        Use QComboBox.currentData() to retrieve the currently selected
        layer.
        The combo box will be cleared as part of this call.

        Parameters
        ----------
        combo_box : QComboBox
            The combo box to add the vector layers to
        """

        def filter_(layer: QgsMapLayer) -> bool:
            if (layer.type() != QgsMapLayerType.VectorLayer
                    or layer.geometryType() != QgsWkbTypes.PolygonGeometry):
                return False
            return has_z is None or has_z == layer_is_3d(layer)

        icon = QgsApplication.getThemeIcon('/mIconPolygonLayer.svg')
        add_layers(self.iface.mapCanvas(), combo_box, *args, icon=icon,
                   filter_=filter_)

    def add_polyline_layers(self, combo_box: QComboBox, *args: QComboBox,
                            has_z: Optional[bool] = None) -> None:
        """Populate the combo box with polyline vector layers.

        Use QComboBox.currentData() to retrieve the currently selected
        layer.
        The combo box will be cleared as part of this call.

        Parameters
        ----------
        combo_box : QComboBox
            The combo box to add the vector layers to
        """

        def filter_(layer: QgsMapLayer) -> bool:
            if (layer.type() != QgsMapLayerType.VectorLayer
                    or layer.geometryType() != QgsWkbTypes.LineGeometry):
                return False
            return has_z is None or has_z == layer_is_3d(layer)

        icon = QgsApplication.getThemeIcon('/mIconLineLayer.svg')
        add_layers(self.iface.mapCanvas(), combo_box, *args, icon=icon,
                   filter_=filter_)

    def add_raster_layers(self, combo_box: QComboBox, *args: QComboBox) -> None:
        """Populate the combo box with raster layers.

        Use QComboBox.currentData() to retrieve the currently selected
        layer.
        The combo box will be cleared as part of this call.

        Parameters
        ----------
        combo_box : QComboBox
            The combo box to add the raster layers to
        """
        icon = QgsApplication.getThemeIcon('/mIconRasterLayer.svg')
        add_layers(self.iface.mapCanvas(), combo_box, *args, icon=icon,
                   filter_=lambda x: x.type()  # type: ignore
                   == QgsMapLayerType.RasterLayer)

    def browse_for_output(self, file_format: str, line_edit: QLineEdit,
                          name_suffix: Optional[str] = None,
                          caption: str = 'Select output file'
                          ) -> Callable[[], None]:
        """Return the callback for an output file picker button.

        This returns a callable that can be passed to the browse
        button's "clicked" slot. This allows customization of the
        callback without requiring an intermediate method.

        Parameters
        ----------
        file_format : str
            The file extension to use
        line_edit : QLineEdit
            The line edit to paste the result string into
        name_suffix : str, optional
            The auto-generated file name to use, by default None
        caption : str, optional
            The caption displayed in the title bar of the file browser,
            by default 'Select output file'

        Returns
        -------
        Callable[[], None]
            Browse button callback
        """
        file_format = file_format.strip('.')
        # The following names will be accessible from within the returned
        # callback
        filter_ = _filter_from_format(file_format)

        # Generate default file name
        filename = get_project_title()
        if name_suffix is not None:
            filename += f'_{name_suffix}'
        filename += f'.{file_format}'
        file_path = os.path.join(get_project_dir(), filename)

        def callback() -> None:
            line_edit.clear()
            file_dialog = QFileDialog()
            file_dialog.setDefaultSuffix(file_format)
            # NOTE: The second tuple element returned is the file filter used
            output_path, _ = file_dialog.getSaveFileName(
                self, caption, file_path, filter_)
            if output_path == '':
                return

            # Add the file extension if it does not exist
            if not output_path.endswith(f'{file_format}'):
                output_path += f'.{file_format}'
            line_edit.setText(output_path)

        return callback

    def print_error(self, err: BaseException) -> None:
        """Display an error message to the user.

        If the error provided is an instance of InterruptedError, this
        does nothing.

        Parameters
        ----------
        err : BaseException
            The error to display
        """
        if err is None:
            return
        # Ignore user interruption signals
        if isinstance(err, InterruptedError):
            return
        text = ('The BASEmesh worker raised an error.\n'
                'See details for traceback and additional information.')
        msg_box = QMessageBox(QMessageBox.Critical, 'BASEmesh - Error', text,
                              buttons=QMessageBox.Ok, parent=self)
        if err.args:
            err_msg = f'{err.__class__.__name__}: '
            err_msg += ', '.join(str(c) for c in err.args)
            msg_box.setInformativeText(err_msg)
        msg_box.setDetailedText(''.join(traceback.format_exception(
            type(err), err, err.__traceback__)))
        msg_box.exec()

    def export_log(self, prefix: str,
                   text_box: QPlainTextEdit) -> Callable[[], None]:
        """Generate and return the log export button's callback.

        The prefix will be used as the default name for the output
        file, together with the timestamp.
        If the prefix were "elevmesh", the default output filename
        would become "log_elevmesh_YYY-MM-DD_HH-MM-SS.log".

        Parameters
        ----------
        prefix : str
            The prefix to prepend to the log timestamp
        text_box : QPlainTextEdit
            The log window whos contents will be saved to a file

        Returns
        -------
        Callable[[], None]
            The export log button callback
        """
        file_dialog = QFileDialog()
        file_dialog.setDefaultSuffix('log')

        def callback() -> None:
            timestamp = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
            base_name = f'log_{prefix}_{timestamp}'
            file_path = os.path.join(get_project_dir(), base_name)

            output_path, _ = file_dialog.getSaveFileName(
                self, 'Save log to file', file_path, 'Text file (*.log)')
            if output_path == '':
                return

            text = text_box.toPlainText()
            # Add a final newline if missing
            if not text.endswith('\n'):
                text += '\n'
            with open(output_path, 'w') as out_file:
                out_file.write(text)

        return callback

    def load_help(self, text_edit: QTextEdit) -> None:
        """Populate the help frame with the corresponding HTML text.

        The text will be generated using a markdown file specified by
        the dialog's help_file attribute.

        Parameters
        ----------
        text_edit : QTextEdit
            The help frame to populate

        Raises
        ------
        AttributeError
            Raised if the help_file attribute has not been set
        """
        try:
            help_file = self.help_file
        except AttributeError as err:
            raise AttributeError(f'{self.__class__.__name__} does not provide '
                                 'a help_file attribute') from err
        help_file = os.path.join(
            get_plugin_dir(), 'plugin', 'gui', 'templates', help_file)
        try:
            with open(help_file) as help_markdown:
                html_str: str = markdown2.markdown(help_markdown.read())
        except FileNotFoundError:
            text_edit.setText(
                f'Unable to generate help HTML, file not found:\n{help_file}')
        else:
            text_edit.setHtml(html_str)

    def run_worker(self, *args: Any, log: Optional[QPlainTextEdit] = None,
                   after: Callable[[Any, Optional[BaseException]], None],
                   cancel_btn: QPushButton, **kwargs: Any) -> None:
        """Run the dialog's "worker" method in another thread.

        Any anonymous arguments and keyword arguments passed to this
        function will be passed on to the worker method. Additionally,
        the worker method's first argument will be a Feedback object
        used to communicate with its parent thread.

        Parameters
        ----------
        after : Callable[[Any, Optional[BaseException]], None]
            A function to run after the worker thread exist, the first
            argument is the return value, the second an exception that
            was raised, if any
        cancel_btn : QPushButton
            A button that will signal the worker function to quit
        log : QPlainTextEdit, optional
            An optional log window to display log messages in via the
            feedback object, by default None
        """
        # NOTE: The following functions are callbacks to run based on worker
        # thread status:

        def cancelled() -> None:
            # Run if cancellation was requested by the user
            worker.cancel()

        def done(return_value: Any) -> None:
            # Run when the worker has finished execution
            teardown()
            after(return_value, None)

        def error(err: BaseException) -> None:
            # Run when the worker thread encountered an exception
            teardown()
            after(None, err)

        def print_(deque_: Deque[str]) -> None:
            # Run when the thread wants to print to stdout
            if log is not None:
                # Consolidate all stdout prints into a single block of text to
                # be inserted at once
                merged = ''.join(deque_)
                deque_.clear()
                cursor = log.textCursor()
                log.insertPlainText(merged)
                log.moveCursor(cursor.End)

        def teardown() -> None:
            # Destroy the worker thread
            thread.quit()
            thread.wait()

        # Instantiate the worker object
        worker = Worker(target=self.worker, args=args, kwargs=kwargs)
        # Create a new thread to run the worker in
        thread = QThread(self)
        worker.moveToThread(thread)

        # Hook up signals
        cancel_btn.clicked.connect(cancelled)
        thread.started.connect(worker.run)  # type: ignore
        thread.finished.connect(worker.deleteLater)  # type: ignore
        worker.done.connect(done)
        worker.error.connect(error)
        worker.print.connect(print_)
        worker.status.connect(self._push_status)

        thread.start()

        time.sleep(0.01)  # HACK: The thread will not start unless we do this

    def _push_status(self, status_tuple: _FeedbackSignal) -> None:
        """Update the progress bar and status message.

        The only parameter is a tuple consisting of three values:
        The new progress percentage (None if unchanged), the new status
        message (None if unchanged), and a flag to show whether the
        worker is busy.

        Parameters
        ----------
        status_tuple : Tuple[Optional[float], Optional[str], bool]
            A status reporting tuple sent from the worker thread
        """
        progress, msg, busy = status_tuple
        if msg is not None:
            self.status.setText(msg)
        if busy:
            # Setting the progress bar to 0/0 causes it to play the "busy"
            # animation
            self.progressBar.setMaximum(0)
            self.progressBar.reset()
        else:
            # If it is not busy, make sure the progress bar is ready to
            # update
            if self.progressBar.maximum() == 0:
                self.progressBar.setMaximum(100)
                self.progressBar.setValue(0)
        if progress is not None and not busy:
            if progress < 0.0:
                progress = 0.0
            elif progress > 1.0:
                progress = 1.0
            self.progressBar.setValue(int(progress*100))


def _filter_from_format(file_format: str) -> str:
    """Return a file format filter for the given format string.

    Allowed format strings are "txt", "log", "2dm", "dxf", "shp",
    "bmg", and "g01".

    Parameters
    ----------
    file_format : str
        The format string to use

    Returns
    -------
    str
        The Qt format string used for file selection
    """
    # The following dict maps file extensions to their canonical format names.
    # It acts as a helper to ensure consistency between tools.
    conversion_dict = {'txt': 'Text file',
                       'log': 'Text file',
                       '2dm': 'SMS 2D mesh',
                       'dxf': 'AutoCAD DXF',
                       'shp': 'Shapefile',
                       'bmg': 'BMG file',
                       'g01': 'HECRAS geometry'}
    name = conversion_dict[file_format]
    return f'{name} (*{file_format})'



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Utilities related to the processing submodule."""

from typing import Optional


def parse_docstring(string: Optional[str]) -> str:
    """Convert a Python docstring for use as an algorithm help text.

    This ignores any leading and trailing whitespace and only keeps
    double linebreaks.

    Parameters
    ----------
    string : Optional[str]
        A docstring to parse into a QGIS help text

    Returns
    -------
    str
        The formatted docstring, or an empty string if no docstring was
        passed
    """
    out_string = ''
    if string is not None:
        for line in string.splitlines():
            line = line.strip()

            # Keep empty lines
            if line == '':
                out_string += '\n\n'
            # Any other lines get appended as normal
            else:
                out_string += f' {line}'
    return out_string



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Custom exceptions for BASEmesh."""


class TopologyError(BaseException):
    """Raised for topology violations."""


class TokenError(Exception):
    """Exception raised for token mismatches.

    This is raised when a given opening token is not matched, or if a
    closing token is encountered without a matchign opening token.
    """



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Custom typing shorthands."""

# NOTE: These shorten type hints and make it easy to find all references to a
# particular type if they need to be updated or replaced.

from typing import Tuple

Point2D = Tuple[float, float]
Point3D = Tuple[float, float, float]
Line2D = Tuple[Point2D, Point2D]
Line3D = Tuple[Point3D, Point3D]
LineString2D = Tuple[Point2D, ...]
LineString3D = Tuple[Point3D, ...]
Polygon2D = Tuple[Point2D, ...]
Polygon3D = Tuple[Point3D, ...]
Rectangle2D = Tuple[float, float, float, float]  # min_x, max_x, min_y, max_y
Triangle2D = Tuple[Point2D, Point2D, Point2D]
Triangle3D = Tuple[Point3D, Point3D, Point3D]



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Shared algorithms which could be reused or optimised.

This module dynamically retrieves algorithms from the C acceleration
module if available.
"""

import os
import sys
import warnings
from .algorithms_py import (counting_sort, dist_2d, dist_3d,
                            distance_to_polygon, get_intersections,
                            half_plane_distance, interpolate_line,
                            interpolate_triangle, line_intersection,
                            line_segments_intersection,
                            point_in_polygon_concave, point_in_polygon_convex,
                            point_on_line, point_within_range,
                            rectangle_intersection, rotate_2d, split_line)

__all__ = [
    'counting_sort',
    'dist_2d',
    'dist_3d',
    'distance_to_polygon',
    'get_intersections',
    'half_plane_distance',
    'interpolate_line',
    'interpolate_triangle',
    'line_intersection',
    'line_segments_intersection',
    'point_in_polygon_concave',
    'point_in_polygon_convex',
    'point_on_line',
    'point_within_range',
    'rectangle_intersection',
    'rotate_2d',
    'split_line'
]

# NOTE: The potential name-shadowing from the wildcard import is by design. In
# fact, it is *only* supposed to overshadow the previously imported names.
# This solution still allows for introspection and offers clear references
# during development while making addition and removal of C implementations
# trivial.
#
# The developer of the C extensions is responsible for ensuring they are a
# perfect match for the native Python implementations.

sys.path.append(os.path.dirname(__file__))
try:
    # pylint: disable=import-error
    from algorithms_c import *  # type: ignore
except ModuleNotFoundError:
    warnings.warn('No compatible C extensions found, using the slower '
                  'pure-Python implementation')



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Channel factory definitions."""

import abc
from typing import Any, List

from .geometry import ChannelGeometry, CrossSection, Vertex

__all__ = [
    'AbstractChannelFactory',
    'TrapezoidalChannel',
    'TrapezoidalChannelWidening'
]


class AbstractChannelFactory(metaclass=abc.ABCMeta):
    """ABC for channel geometry factories."""

    @abc.abstractmethod
    def build_channel(self) -> ChannelGeometry:
        """Build a channel geometry using whatever inputs were given."""
        ...


class TrapezoidalChannel(AbstractChannelFactory):
    """Generate a straight channel with trapezoidal cross sections.

    Attributes
    ----------
    dist_cs : float
        Distance between two cross sections in the channel.
    num_cs : int
        Number of cross sections in the channel
    bed_width : float
        The width of the channel bed.
    height : float
        The total height of the channel. This, together with
        :attr:`bank_slope`, controls the total width of the channel.
    bed_slope : float
        The rise-over-run slope of the channel.
    bank_slope : float
        The rise-over-run slope of the channel banks. This, together
        with :attr:`height`, controls the total width of the channel.
    friction_bed : float
        Strickler friction value for the channel bed.
    friction_bank : float
        Strickler friction value for the channel banks.
    midpoint_shift : float, optional
        Relative position of the nodes located in the middle of the
        channel banks. Lower values move these nodes closer to the
        channel bed, higher values move them towards the outer edge of
        the channel.
        Allowed values are 0.0 to 1.0 (exclusive). Defaults to 0.5.
    cross_sections : int, optional
        The number of cross sections to generate. Defaults to 20.
    raise_by : float, optional
        The starting elevation of the channel origin. Use this if you
        want to keep elevations positive. Defaults to 0.

    """

    def __init__(self, dist_cs: float, num_cs: int, bed_width: float,
                 height: float, bed_slope: float, bank_slope: float,
                 friction_bed: float, friction_bank: float,
                 midpoint_shift: float = 0.5, raise_by: float = 0.0) -> None:
        self.dist_cs = dist_cs
        self.num_cs = num_cs
        self.bed_width = bed_width
        self.height = height
        self.bed_slope = bed_slope
        self.bank_slope = bank_slope
        self.friction_bed = friction_bed
        self.friction_bank = friction_bank
        self.midpoint_shift = midpoint_shift
        self.raise_by = raise_by
        self._validate_parameters()

    @property
    def cs_total(self) -> int:
        """Return the total number of cross sections in the channel."""
        return self.num_cs

    def build_channel(self) -> ChannelGeometry:
        """Build the channel geometry."""
        # Validate the attributes again in case the user updated them since
        # instantiation of the factory
        self._validate_parameters()
        # Generate the channel geometry
        channel = ChannelGeometry()
        for index in range(self.num_cs):
            cs_ = self._cross_section_factory(index+1, self.dist_cs*index)
            # Apply Z offset
            cs_.reference_height += self.raise_by
            # Define frictions
            cs_.friction_slices = [(1, 2), (3, 4), (5, 6)]
            cs_.friction_coefficients = [
                self.friction_bank, self.friction_bed, self.friction_bank]
            channel.add_cross_section(cs_)
        return channel

    def _cross_section_factory(self, name: Any,
                               distance: float) -> CrossSection:
        """Return a cross section by distance from the origin.

        Parameters
        ----------
        name : Any
            The unique name of the cross section. The given value will
            be cast to string; it is safe to pass the current cross
            section index into this parameter.
        distance : float
            The distance of this cross section from the origin. This
            acts much like the :attr:`CrossSection.distance_coord`
            attribute, but is provided in metres.

        Returns
        -------
        CrossSection
            The cross section to insert at the given distance.

        """
        # NOTE: Sketch of how these vertices are arranged (by index).
        # These indices also correspond to the constraint applied to each
        # vertex (i.e. the matching nodes will be connected by break lines.)
        #
        # 0               6
        #  \             /  * Note that there is no breakline
        #   1           5     between the "3"-nodes
        #    \         /
        #     2---3---4
        #
        half_bed_width = 0.5 * self.bed_width
        half_channel_width = half_bed_width + self.height/self.bank_slope
        half_bank_width = half_channel_width - half_bed_width

        # Calculate vertices
        vertices = [
            # Left top
            Vertex(-half_channel_width, self.height, 0),
            # Left mid
            Vertex(-half_bed_width - self.midpoint_shift*half_bank_width,
                   self.midpoint_shift * self.height, 1),
            # Left bot
            Vertex(-half_bed_width, 0.0, 2),
            # Middle / flow axis
            Vertex(0.0, 0.0),
            # Right bot
            Vertex(half_bed_width, 0.0, 4),
            # Right mid
            Vertex(half_bed_width + self.midpoint_shift*half_bank_width,
                   self.midpoint_shift * self.height, 5),
            # Right top
            Vertex(half_channel_width, self.height, 6)]

        # Adjust vertices based on slope
        for vertex in vertices:
            pos_y, pos_z = vertex.pos
            pos_z += distance * self.bed_slope
            vertex.pos = pos_y, pos_z

        # Create cross section
        cs_ = CrossSection(str(name), vertices)
        cs_.anchor = distance, -(cs_.length / 2.0)
        cs_.flow_axis_coord = distance
        return cs_

    def _validate_parameters(self) -> None:
        """Validate the factory's attributes.

        This raises ValueErrors for any invalid arguments. This is done
        in a separate method as this must be re-run when building the
        channel as the user might have updated some values since the
        factory was instantiated.
        """
        if self.dist_cs < 0.0:
            raise ValueError(
                'Inter-cross-sectional distance must be greater than zero')
        if self.bed_width <= 0.0:
            raise ValueError('Channel bed width must be greater than zero')
        if self.height <= 0.0:
            raise ValueError('Channel height must be greater than zero')
        if not 0.0 < self.midpoint_shift < 1.0:
            raise ValueError(
                'Midpoint shift must be greater than 0.0 and less than 1.0')
        if self.num_cs < 2:
            raise ValueError('At least two cross sections required')


class TrapezoidalChannelWidening(TrapezoidalChannel):
    r"""Generate a straight trapezoidal channel with a widened section.

    This is a subclass of :class:`TrapezoidalChannel`; please refer to
    its documentation for information shared attributes.
    Note that the minimum number of sections require is six for this
    class, instead of the two required for the regular channel. Four of
    these are defined in pairs for each of the straight channel
    sections (i.e. `num_cs >= 2`), the other two are defined for the
    widened section through `widening_num_cs`.

    `transition_num_cs` is allowed to be 0, in which case no cross
    sections wil lbe inserted between the regular and widened sections.
    Note that `transition_dist_cs` is still used and may not be 0 in
    this case as it controls the distance between the ther sections.

    Attributes
    ----------
    widening_dist_cs : float
        Distance between two cross sections in the widened section.
    widening_num_cs : int
        Number of cross sections in the widenined section.
    transition_dist_cs : float
        Distance between two cross sections in the transition. If
        `transition_num_cs` is 0, this controls the distance between
        the other sections at the step.
    transition_num_cs : int
        Number of cross sections to insert at the transition. May be 0.
    widening_bed_width : float
        The bed width for the widened section.
    friction_widening : float
        Strickler friction value for banks in the widened section.

    """

    def __init__(self, dist_cs: float, num_cs: int, bed_width: float,
                 height: float, bed_slope: float, bank_slope: float,
                 friction_bed: float, friction_bank: float,
                 widening_bed_width: float, widening_friction: float = 30.0,
                 widening_dist_cs: float = 30.0, widening_num_cs: int = 6,
                 transition_dist_cs: float = 20.0, transition_num_cs: int = 3,
                 midpoint_shift: float = 0.5, raise_by: float = 0.0) -> None:
        # Initialise TrapezoidalChannel
        super().__init__(dist_cs, num_cs, bed_width, height, bed_slope,
                         bank_slope, friction_bed, friction_bank,
                         midpoint_shift=midpoint_shift, raise_by=raise_by)
        # Add custom attributes
        self.widening_bed_width = widening_bed_width
        self.widening_friction = widening_friction
        self.widening_dist_cs = widening_dist_cs
        self.widening_num_cs = widening_num_cs
        self.transition_dist_cs = transition_dist_cs
        self.transition_num_cs = transition_num_cs
        # Validate parameters (this also calls the parent class's validator)
        self.__validate_parameters()

    @property
    def cs_total(self) -> int:
        """Return the total number of cross sections in the channel."""
        return (self.num_cs * 2 + self.transition_num_cs * 2
                + self.widening_num_cs)

    @property
    def _transition_start(self) -> float:
        """Return the starting distance for the first transition."""
        return (self.num_cs-1) * self.dist_cs

    @property
    def _transition_end(self) -> float:
        """Return the starting distance for the second transition"""
        return (self._widening_start
                + self.widening_dist_cs * (self.widening_num_cs-1))

    @property
    def _widening_start(self) -> float:
        """Return the starting distance for the widened section."""
        # NOTE: +1 is correct; -1 plus 2 from the global transition CS offset
        return (self._transition_start
                + self.transition_dist_cs * (self.transition_num_cs+1))

    def build_channel(self) -> ChannelGeometry:
        """Build the channel geometry."""
        # Validate the attributes again in case the user updated them since
        # instantiation of the factory
        self._validate_parameters()

        cross_sections: List[CrossSection] = []
        total_index = 1
        total_offset = 0.0

        # Add initial straight section
        for index in range(self.num_cs):
            offset = total_offset + index * self.dist_cs
            section = self._cross_section_factory(total_index+index, offset)
            # Define frictions
            section.friction_slices = [(1, 2), (3, 4), (5, 6)]
            section.friction_coefficients = [self.friction_bank,
                                             self.friction_bed,
                                             self.friction_bank]
            cross_sections.append(section)
        total_offset += self.dist_cs * (self.num_cs - 1)
        total_index += self.num_cs

        # Add transitioning
        for index in range(1, self.transition_num_cs+1):
            offset = total_offset + index * self.transition_dist_cs
            section = self._cross_section_factory(total_index+index, offset)
            # Define frictions
            section.friction_slices = [(1, 2), (3, 4), (5, 6)]
            section.friction_coefficients = [self.widening_friction,
                                             self.friction_bed,
                                             self.widening_friction]
            cross_sections.append(section)
        # NOTE: +1 is intended: +2 from the global offset, then -1
        total_offset += self.transition_dist_cs * (self.transition_num_cs + 1)
        total_index += self.transition_num_cs

        # Add widening
        for index in range(self.widening_num_cs):
            offset = total_offset + index * self.widening_dist_cs
            section = self._cross_section_factory(total_index+index, offset)
            # Define frictions
            section.friction_slices = [(1, 2), (3, 4), (5, 6)]
            section.friction_coefficients = [self.widening_friction,
                                             self.friction_bed,
                                             self.widening_friction]
            cross_sections.append(section)
        total_offset += self.widening_dist_cs * (self.widening_num_cs - 1)
        total_index += self.widening_num_cs

        # Add second transitioning
        for index in range(1, self.transition_num_cs+1):
            offset = total_offset + index * self.transition_dist_cs
            section = self._cross_section_factory(total_index+index, offset)
            # Define frictions
            section.friction_slices = [(1, 2), (3, 4), (5, 6)]
            section.friction_coefficients = [self.widening_friction,
                                             self.friction_bed,
                                             self.widening_friction]
            cross_sections.append(section)
        # NOTE: +1 is correct, see first transition
        total_offset += self.transition_dist_cs * (self.transition_num_cs + 1)
        total_index += self.transition_num_cs

        # Add second straight sections
        for index in range(self.num_cs):
            offset = total_offset + index * self.dist_cs
            section = self._cross_section_factory(total_index+index, offset)
            # Define frictions
            section.friction_slices = [(1, 2), (3, 4), (5, 6)]
            section.friction_coefficients = [self.friction_bank,
                                             self.friction_bed,
                                             self.friction_bank]
            cross_sections.append(section)

        # Apply Z offset
        for index, section in enumerate(cross_sections):
            section.reference_height = self.raise_by

        # Return channel geometry
        return ChannelGeometry(cross_sections=cross_sections)

    def _cross_section_factory(self, name: Any,
                               distance: float) -> CrossSection:
        # NOTE: This largely mirrors the factory in the regular
        # TrapezoidalChannel factory, refer to it for details. This version
        # extends its approach by interpolating the appropriate bed width based
        # on position along the channel (widening or not, transitions, etc.).

        # Get widening factor
        transition_length = self._widening_start - self._transition_start
        widening_rel: float
        if distance <= self._transition_start:
            widening_rel = 0.0
        elif distance < self._widening_start:
            widening_rel = (
                distance - self._transition_start) / transition_length
        elif distance <= self._transition_end:
            widening_rel = 1.0
        elif distance < self._transition_end + transition_length:
            widening_rel = abs(1 - (
                distance - self._transition_end) / transition_length)
        else:
            widening_rel = 0.0

        # Create cross section
        half_bed_width = self.bed_width*0.5 + (
            (self.widening_bed_width - self.bed_width) * widening_rel * 0.5)
        half_channel_width = half_bed_width + self.height/self.bank_slope
        half_bank_width = half_channel_width - half_bed_width
        # Calculate vertices
        vertices = [
            # Left top
            Vertex(-half_channel_width, self.height, 0),
            # Left mid
            Vertex(-half_bed_width - self.midpoint_shift*half_bank_width,
                   self.midpoint_shift * self.height, 1),
            # Left bot
            Vertex(-half_bed_width, 0.0, 2),
            # Middle / flow axis
            Vertex(0.0, 0.0),
            # Right bot
            Vertex(half_bed_width, 0.0, 4),
            # Right mid
            Vertex(half_bed_width + self.midpoint_shift*half_bank_width,
                   self.midpoint_shift * self.height, 5),
            # Right top
            Vertex(half_channel_width, self.height, 6)]

        # Create cross section
        cs_ = CrossSection(str(name), vertices)
        cs_.anchor = distance, -(cs_.length / 2.0)
        cs_.flow_axis_coord = distance
        cs_.reference_height = distance * self.bed_slope
        return cs_

    def __validate_parameters(self) -> None:
        """Validate the factory's attributes.

        This raises ValueErrors for any invalid arguments. This is done
        in a separate method as this must be re-run when building the
        channel as the user might have updated some values since the
        factory was instantiated.
        """
        if self.widening_dist_cs <= 0.0:
            raise ValueError(
                'Inter-cross-sectional distance must be greater than zero')
        if self.widening_num_cs <= 0:
            raise ValueError(
                'The widening must have at least two cross sections')
        if self.transition_dist_cs <= 0.0:
            raise ValueError(
                'Inter-cross-sectional distance must be greater than zero')
        if self.widening_bed_width <= 0.0:
            raise ValueError('Widening bed width must be greater than zero')
        if self.widening_num_cs < 2:
            raise ValueError('Number of cross sections in the widening must '
                             'be at least 2')
        if self.transition_num_cs < 0:
            raise ValueError('Invalid number of cross sections for transition')
        super()._validate_parameters()



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Python implementation of the mesh interpolation GUI."""

import os
from typing import Any, List, Optional, Tuple, Union
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (QCheckBox, QComboBox, QGroupBox, QLineEdit,
                             QListWidget, QListWidgetItem, QPushButton,
                             QRadioButton, QTabWidget, QTextEdit)
from qgis.core import (QgsMapLayerType, QgsMeshLayer,  # type: ignore
                       QgsProject, QgsRasterLayer)
from qgis.gui import QgisInterface  # type: ignore
from ...abc import ElevationSource
from ...core import Mesh
from ...feedback import Feedback
from ...interpolation import calculate_element_elevation, interpolate_mesh
from ..raster import RasterDataWrapper
from .base import DialogBase

# This is a custom type hint used to refer to mesh layers and tuples of a
# raster layer and its corresponding band index.
_ElevLayer = List[Union[QgsMeshLayer, Tuple[QgsRasterLayer, int]]]


class InterpolDialog(DialogBase):
    """GUI implementaiton of the interpolation utility."""

    # NOTE: Qt widgets just have a ton of attributes - too bad!
    # pylint: disable=too-many-instance-attributes
    # Qt also uses camelCase for its attributes - too bad!
    # pylint: disable=invalid-name

    ui_file = 'ui_interpol.ui'
    help_file = 'help_interpol.md'

    def __init__(self, iface: QgisInterface) -> None:
        """Connect signal callbacks and populate widgets."""
        super().__init__(iface)  # Initialize DialogBase

        # NOTE: The following are attribute type hints used to inform supported
        # IDEs of the attribute types defined in the Qt UI file. This code is
        # executed during runtime, but does not do anything as far as the
        # Python interpreter is concerned.
        self.inputGroup: QGroupBox
        self.layer: QComboBox
        self.tabWidget: QTabWidget
        self.useMeshSource: QRadioButton
        self.meshSource: QComboBox
        self.useRasterSource: QRadioButton
        self.rasterSource: QComboBox
        self.rasterBand: QComboBox
        self.selectedList: QListWidget
        self.availableList: QListWidget
        self.help: QTextEdit
        self.outputGroup: QGroupBox
        self.format: QComboBox
        self.filePath: QLineEdit
        self.fileBrowseBtn: QPushButton
        self.addToMap: QCheckBox
        self.cancelBtn: QPushButton
        self.runBtn: QPushButton
        self.closeBtn: QPushButton

        self._use_advanced = False
        self.load_help(self.help)

        # Populate list widgets
        self._layers_cache: _ElevLayer = []
        self._set_available_layers()  # Populate list widgets

        # Populate combo boxes
        self.add_mesh_layers(self.layer, self.meshSource)
        self.add_raster_layers(self.rasterSource)

        # Hook up callbacks
        self.tabWidget.currentChanged.connect(self._tab_switched)
        self.rasterSource.currentIndexChanged.connect(
            self._update_raster_bands)
        self.fileBrowseBtn.clicked.connect(
            self.browse_for_output('2dm', self.filePath, 'computational-mesh'))

        # Manually trigger some callbacks to pre-populate the combo boxes
        self.rasterSource.currentIndexChanged.emit(0)

    def accept(self) -> None:
        """Evaluate the user's chosen settings and run the tool.

        This function runs when the user confirms their settings, its
        name may not be changed as it is tied directly into the default
        QDialog slot.
        """
        # Input layer
        target_layer = self.layer.currentData()

        # NOTE: The following list contains the elevation sources to be used
        # for interpolation in descending order of priority.
        sources: List[ElevationSource] = []

        # Advanced mode
        if self._use_advanced:
            for list_item in self._get_list_selection():
                # Mesh layer
                if isinstance(list_item, QgsMeshLayer):
                    layer: QgsMeshLayer = list_item
                    filename = layer.source()
                    sources.append(Mesh.open(filename))
                # Raster layer
                else:
                    layer, band_index = list_item
                    sources.append(RasterDataWrapper(layer, band_index))

        # Basic mode
        else:
            # Use elevation mesh
            if self.useMeshSource.isChecked():
                layer = self.meshSource.currentData()
                sources = [Mesh.open(layer.source())]
            # Use raster data
            else:
                layer = self.rasterSource.currentData()
                band_index = self.rasterBand.currentData()
                sources = [RasterDataWrapper(layer, band_index)]

        # Output
        output_path = self.filePath.text()
        format_value = self.format.currentIndex()
        output_bm2 = output_bm3 = False
        if format_value != 1:
            output_bm2 = True
        if format_value >= 1:
            output_bm3 = True
        add_file_to_map = self.addToMap.isChecked()

        def after(_: Any, error: Optional[BaseException]) -> None:
            """Run after the worker has finished.

            Parameters
            ----------
            _ : Any
                The worker's return value; discarded and unneeded
            error : Optional[BaseException]
                The error that forced the end of execution, by default
                None
            """
            # Add file to map
            if error is None and add_file_to_map:
                name, _ = tuple(os.path.basename(output_path).rsplit('.', 1))
                layer = QgsMeshLayer(output_path, name, 'mdal')
                QgsProject.instance().addMapLayer(layer)
            self._set_lockout(False)
            if error is not None:
                self._push_status((0.0, 'Error', False))
                self.print_error(error)

        self._set_lockout(True)
        self.run_worker(target_layer, sources, output_path, output_bm2,
                        output_bm3, after=after, cancel_btn=self.cancelBtn)

    @staticmethod
    def worker(feedback: Feedback, target_layer: QgsMeshLayer,
               sources: List[_ElevLayer], output_path: str,
               output_bm2: bool, output_bm3: bool) -> None:
        """Perform the operation triggered by the dialog.

        This method is static as it will be carried out by a separate
        thread.

        Parameters
        ----------
        feedback : Feedback
            The feedback object use to communicate with the parent
            thread; this will be auto-filled internally
        target_layer : QgsMeshLayer
            The mesh layer to interpolate
        sources : List[ElevationSource]
            The elevation sources to use for interpolation
        output_path : str
            The output path of the interpolated mesh
        output_bm2 : bool
            Whether to interpolate nodes (BASEMENT v2.8)
        output_bm3 : bool
            Whether to interpolate elements (BASEMENT v3.x)

        Raises
        ------
        ValueError
            Raised if neither output format is True
        """
        if not any((output_bm2, output_bm3)):
            raise ValueError('At least one output format required')

        # Load input mesh file
        with feedback.busy('Loading quality mesh...'):
            interpol_mesh = Mesh.open(target_layer.source())

        # Interpolate nodes
        if output_bm2:
            if output_bm3:
                feedback.set_scaling(0.5)
            feedback.update(0.0, 'Interpolating mesh nodes...')
            interpolate_mesh(
                interpol_mesh, *sources, feedback=feedback)  # type: ignore

        # Interpolate elements
        if output_bm3:
            if output_bm2:
                feedback.set_scaling(0.5, 0.5)
            feedback.update(0.0, 'Interpolating mesh elements...')
            # Calculate element elevations
            elevations = calculate_element_elevation(
                interpol_mesh, *sources, feedback=feedback)  # type: ignore
            # Update element materials
            for element in interpol_mesh.elements:
                element.materials = *element.materials, elevations[element.id_]

        # Save output mesh file
        with feedback.busy('Saving computational mesh...'):
            interpol_mesh.save(output_path, 2 if output_bm3 else 1)

        feedback.clear_scaling()
        feedback.update(1.0, 'Done')

    def _get_list_selection(self) -> _ElevLayer:
        """Return the selected layers from the "Advanced" tab lists."""
        return_list = []
        for index in range(self.selectedList.count()):
            item = self.selectedList.item(index)
            layer_index: int = item.data(Qt.UserRole)
            return_list.append(self._layers_cache[layer_index])
        return return_list

    def _set_available_layers(self) -> None:
        """Add all elevation sources to the available layers list."""
        # Clear the lists
        avail_list = self.availableList
        avail_list.clear()
        self.selectedList.clear()
        self._layers_cache = []
        # Add all mesh layers, as well as any raster bands
        for layer in self.iface.mapCanvas().layers():
            if layer.type() == QgsMapLayerType.MeshLayer:
                if not layer.source().endswith('.2dm'):
                    # NOTE: Only 2DM meshes supported for now
                    continue
                new_item = QListWidgetItem(layer.name())
                new_item.setData(Qt.UserRole, len(self._layers_cache))
                self._layers_cache.append(layer)
                avail_list.addItem(new_item)
            elif layer.type() == QgsMapLayerType.RasterLayer:
                for index in range(layer.bandCount()):
                    band_name = layer.bandName(index)
                    new_item = QListWidgetItem(f'{layer.name()} • {band_name}')
                    new_item.setData(Qt.UserRole, len(self._layers_cache))
                    self._layers_cache.append((layer, index))
                    avail_list.addItem(new_item)

    def _set_lockout(self, is_locked: bool) -> None:
        """Lock out the GUI while running the tool.

        Parameters
        ----------
        is_locked : bool
            Whether to lock or unlock the GUI
        """
        self.inputGroup.setEnabled(not is_locked)
        self.tabWidget.setEnabled(not is_locked)
        self.outputGroup.setEnabled(not is_locked)
        self.closeBtn.setEnabled(not is_locked)
        self.runBtn.setEnabled(not is_locked)
        self.cancelBtn.setEnabled(is_locked)

    def _tab_switched(self, tab_index: int) -> None:
        """Switch between simple and advanced tabs.

        Parameters
        ----------
        tab_index : int
            The index of the tab that was switched to
        """
        # Basic tab
        if tab_index == 0 and self._use_advanced:
            self._use_advanced = False
        # Advanced tab
        elif tab_index == 1 and not self._use_advanced:
            self._use_advanced = True

    def _update_raster_bands(self) -> None:
        """Refresh the raster layer band combo box.

        This removes all items from the basic tab's raster band combo
        box and populates it with the bands found for the currently
        selected raster layer.
        """
        raster_layer = self.rasterSource.currentData()
        band_box = self.rasterBand
        band_box.clear()
        if raster_layer is None:
            return
        for index in range(raster_layer.bandCount()):
            band_name = raster_layer.bandName(index)
            band_box.addItem(band_name)
            band_box.setItemData(band_box.count()-1, index)



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Reader and writer functions for triangle input/output formats."""

import datetime
import warnings
from typing import Callable, Collection, Iterable, List, Optional, Set, TypeVar
from .log import logger
from .markers import HoleMarker, RegionMarker
from .pslg import PSLGElement, PSLGNode, PSLGSegment

# Generic type variable used for type-hinting _create_consecutive_list
T = TypeVar('T')


def read_elements(filename: str, nodes: List[PSLGNode],
                  use_matid: bool = True) -> List[PSLGElement]:
    """Read an output ELE file and return a list of elements.

    Parameters
    ----------
    filename : str
        The file to process
    nodes : List[PSLGNode]
        The sorted list of nodes for these elements
    use_matid : bool, optional
        Whether to expect a material ID field, by default True

    Returns
    -------
    List[PSLGElement]
        The list of elements found in the file
    """
    logger.debug('Reading %s', filename)
    with open(filename) as input_file:
        elements: List[PSLGElement] = []
        ele_count: int
        line: str
        for line in input_file:
            line = line.strip()
            # Ignore blank lines
            if line == '':
                continue
            # Ignore comments
            if line.startswith('#'):
                continue
            # File header
            try:
                _ = ele_count
            except NameError:
                ele_count = int(line.split()[0])
                continue
            # File contents
            elements.append(PSLGElement.from_string(
                line, nodes, use_matid=use_matid))
    if len(elements) != ele_count:
        warnings.warn(f'{ele_count} elements expected, got {len(elements)}')
    return elements


def read_nodes(filename: str, attr_names: Optional[List[str]] = None
               ) -> List[PSLGNode]:
    """Read an output NODE file and return a list of nodes.

    Parameters
    ----------
    filename : str
        The file to process
    attr_names : List[str], optional
        If node attributes were specified before triangulation, you can
        use this list to restore the original attribute names, by
        default None

    Returns
    -------
    List[PSLGNode]
        The list of nodes found in the file
    """
    logger.debug('Reading %s', filename)
    with open(filename) as input_file:
        nodes: List[PSLGNode] = []
        node_count: int
        use_boundary: bool

        line: str
        for line in input_file:
            line = line.strip()
            # Ignore blank lines
            if line == '':
                continue
            # Ignore comments
            if line.startswith('#'):
                continue
            # File header
            try:
                _ = node_count
            except NameError:
                words = line.split()
                node_count = int(words[0])
                use_boundary = bool(int(words[3]))
                continue
            # File contents
            nodes.append(PSLGNode.from_string(
                line, use_boundary=use_boundary, column_names=attr_names))
    if len(nodes) != node_count:
        warnings.warn(f'{node_count} nodes expected, got {len(nodes)}')
    return nodes


def write_nodes(filename: str, nodes: Collection[PSLGNode], *,
                ignore_existing_ids: bool = True,
                node_attr_keys: Optional[List[str]] = None) -> None:
    """Write an input NODE file from a set of input nodes.

    By default, this function will assign random node IDs upon writing.
    You can prevent this by setting the ignore_existing_ids kwarg to
    False, which will cause an error if the highest ID exceeds the
    number of provided nodes. This is due to "Triangle" expecting
    consecutive ID numbering without jumps, which cannot be achieved in
    this case.

    Parameters
    ----------
    filename : str
        The file to write
    nodes : Collection[PSLGNode]
        The nodes to write to file
    ignore_existing_ids : bool, optional
        Whether to write the existing node IDs, otherwise they will be
        numbered sequentially upon writing, by default True
    node_attr_keys : List[str], optional
        The list of noe attribute keys to write to file, by default
        None

    Raises
    ------
    ValueError
        Raised if the number of nodes is insufficient to generate a
        consecutive list of ascending node IDs. This can only occur if
        ignore_existing_ids is enabled.
    """
    logger.debug('Writing %s', filename)
    # pylint: disable=import-outside-toplevel
    from . import __version__ as module_version
    # Process the input nodes to consolidate all node attributes
    max_id = 0
    for node in nodes:
        if (not ignore_existing_ids
                and node.id_ is not None
                and node.id_ > max_id):
            max_id = node.id_
    # Check node.id_ count
    if nodes and len(nodes) <= max_id:
        raise ValueError('Unable to retain existing IDs: The highest ID '
                         f'encountered ({max_id}) exceeds the number of nodes '
                         f'provided ({len(nodes)})')

    # Write output file
    with open(filename, 'w') as output_file:
        # Custom file header comment
        now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        output_file.write(f'# Triangle NODE file\n# Generated {now} using '
                          f'"BASEmesh Triangle Wrapper" v{module_version}\n\n')
        # Node information header
        attr_count = 0 if node_attr_keys is None else len(node_attr_keys)
        output_file.write('# Vertex count | Dimension | Attribute count | '
                          f'Use boundary markers\n{len(nodes)} 2 '
                          f'{attr_count} 0\n')
        # Node table header
        output_file.write('# Vertex ID | X | Y')
        if node_attr_keys:
            output_file.write(f' | {" | ".join(node_attr_keys)}')
        output_file.write('\n')

        # Node table contents
        if ignore_existing_ids:
            # Use arbitrary node order
            sorted_nodes = list(nodes)
        else:
            # Split nodes into fixed ID and random ID
            sorted_nodes = _create_consecutive_list(nodes)
        for index, node in enumerate(sorted_nodes):
            output_file.write(
                f'{index} {node.as_tuple()[0]} {node.as_tuple()[1]}')
            if node_attr_keys:
                attr_list = []
                for attr in node_attr_keys:
                    try:
                        val = float(node.attributes[attr])
                    except KeyError:
                        attr_list.append('0')
                    except ValueError:
                        # Silently ignore any attributes that cannot be
                        # converted to float
                        pass
                    else:
                        if val.is_integer():
                            val = int(val)
                        attr_list.append(str(val))
                output_file.write(f' {" ".join(attr_list)}')
            output_file.write('\n')


def write_poly(filename: str, *, nodes: Optional[Collection[PSLGNode]] = None,
               segments: Optional[Collection[PSLGSegment]] = None,
               holes: Optional[Collection[HoleMarker]] = None,
               regions: Optional[Collection[RegionMarker]] = None,
               node_attr_keys: Optional[List[str]] = None,
               include_area: bool = True,
               include_matid: bool = True,
               ignore_existing_ids: bool = True) -> None:
    """Write an input POLY file from the given input data.

    As segments are also defined through nodes, they may be sufficient
    for triangulation, even without any explicit nodes specified.
    nodes is therefore a kwarg as well, although the total number of
    nodes (both explicitly defined ones and ones inferred through
    segments) must not be zero.

    By default, this function will assign random object IDs upon
    writing. You can prevent this by setting the ignore_existing_ids
    kwarg to False, which will cause an error if the highest ID
    exceeds the number of provided objects. This is due to "Triangle"
    expecting consecutive ID numbering without jumps, which cannot be
    achieved in this case.

    Parameters
    ----------
    filename : str
        The file to write
    nodes : Collection[PSLGNode], optional
        The nodes to write to file, by default None
    segments : Collection[PSLGSegment], optional
        The segments to write to file, by default None
    holes : Collection[HoleMarker], optional
        Any hole markers to include, by default None
    regions : Collection[RegionMarker], optional
        Any region markers to include, by default None
    node_attr_keys : List[str], optional
        The keys in the node attribute dicts to write, by default None
    include_area : bool, optional
        Whether to write region area markers, by default True
    include_matid : bool, optional
        Whether to write region attributes, by default True
    ignore_existing_ids : bool, optional
        Whether to keep existing node IDs or regenerate them, by
        default True

    Raises
    ------
    ValueError
        Raised if no nodes were specified or inferrable via segments
    ValueError
        Raised if the number of objects was insufficient to keep
        existing node IDs when writing
    """
    logger.debug('Writing %s', filename)
    # pylint: disable=import-outside-toplevel
    from . import __version__ as module_version
    # Process the input segments and merge their nodes with the ones provided
    node_set: Set[PSLGNode] = set()
    if segments is not None:
        for segment in segments:
            node_set.update((segment.start, segment.end))
    if nodes is not None:
        node_set.update(nodes)
    if not node_set:
        raise ValueError('No nodes were specified and none could be inferred '
                         'from segments')
    # Process the nodes to consolidate all node attributes
    max_node_id = 0
    for node in node_set:
        if (not ignore_existing_ids
                and node.id_ is not None
                and node.id_ > max_node_id):
            max_node_id = node.id_
    # Check node.id_ count
    if not ignore_existing_ids and node_set and len(node_set) <= max_node_id:
        raise ValueError('Unable to retain existing node IDs: The highest '
                         f'node ID encountered ({max_node_id}) exceeds the '
                         f'number of nodes provided ({len(node_set)})')
    # Check segment ids
    max_segment_id = 0
    if segments is not None:
        for segment in segments:
            if (not ignore_existing_ids
                    and segment.id_ is not None
                    and segment.id_ > max_segment_id):
                max_segment_id = segment.id_
    if segments and len(segments) <= max_segment_id:
        raise ValueError('Unable to retain existing segment IDs: The highest '
                         f'segment ID encountered ({max_segment_id}) exceeds '
                         f'the number of segments provided ({len(segments)})')

    # Write output file
    with open(filename, 'w') as output_file:
        # Custom file header comment
        now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        output_file.write(f'# Triangle POLY file\n# Generated {now} using '
                          f'"BASEmesh Triangle Wrapper" v{module_version}\n\n')

        # Node information header
        node_key_count = 0 if node_attr_keys is None else len(node_attr_keys)
        output_file.write('# Vertex count | Dimension | Attribute count | '
                          f'Use boundary markers\n{len(node_set)} 2 '
                          f'{node_key_count} 0\n')
        # Node table header
        output_file.write('# Vertex ID | X | Y')
        if node_attr_keys:
            output_file.write(f' | {" | ".join(node_attr_keys)}')
        output_file.write('\n')

        # Node table contents
        if ignore_existing_ids:
            # Use arbitrary node order
            sorted_nodes = list(node_set)
        else:
            sorted_nodes = _create_consecutive_list(node_set)
        for index, node in enumerate(sorted_nodes):
            output_file.write(
                f'{index} {node.as_tuple()[0]} {node.as_tuple()[1]}')
            if node_attr_keys:
                attr_list = []
                for attr in node_attr_keys:
                    try:
                        val = float(node.attributes[attr])
                    except KeyError:
                        attr_list.append('0')
                    except ValueError:
                        # Silently ignore any attributes that cannot be
                        # converted to float
                        pass
                    else:
                        if val.is_integer():
                            val = int(val)
                        attr_list.append(str(val))
                output_file.write(f' {" ".join(attr_list)}')
            output_file.write('\n')

        # Segment information header
        output_file.write('\n# Segment count | Use boundary markers\n')
        if segments:
            output_file.write(str(len(segments)))
        else:
            output_file.write('0')
        output_file.write(' 0\n')
        if segments:
            # Segment table header
            output_file.write('# Segment ID | Start node | End node')
            output_file.write('\n')
            # Segment table contents
            for index, segment in enumerate(segments):
                start_node = sorted_nodes.index(segment.start)
                end_node = sorted_nodes.index(segment.end)
                output_file.write(f'{index} {start_node} {end_node}')
                output_file.write('\n')

        # Hole information header
        output_file.write('\n# Hole count\n')
        if holes:
            output_file.write(str(len(holes)))
        else:
            output_file.write('0')
        output_file.write('\n')

        if holes:
            # Hole table header
            output_file.write('# Hole ID | X | Y\n')
            # Hole table contents
            for index, hole in enumerate(holes):
                output_file.write(
                    f'{index} {hole.as_tuple()[0]} {hole.as_tuple()[1]}\n')

        # Region information header
        output_file.write('\n# Region count\n')
        if regions:
            output_file.write(str(len(regions)))
        else:
            output_file.write('0')
        output_file.write('\n')

        if regions:
            # Region table header
            output_file.write('# Region ID | X | Y')
            if include_matid:
                output_file.write(' | MATID')
            if include_area:
                output_file.write(' | Max area')
            output_file.write('\n')
            # Region table contents
            for index, region in enumerate(regions):
                output_file.write(
                    f'{index} {region.as_tuple()[0]} {region.as_tuple()[1]}')
                if include_matid:
                    output_file.write(f' {region.attribute}')
                if include_area:
                    output_file.write(f' {region.max_area}')
                output_file.write('\n')


def _create_consecutive_list(
        iterable: Iterable[T], *,
        key: Optional[Callable[[T], Optional[int]]] = None) -> List[T]:
    """Process the input list and create a key-sorted copy of it.

    This function will take the elements from the input list and group
    them by the given key, which is expected to be an optional integer.

    It then creates a new list with all non-None key values being
    placed at the appropriate index and the others placed randomly to
    "fill the gaps".

    Parameters
    ----------
    iterable : Iterable[T]
        The iterable to process
    key : Callable[[T], Optional[int]], optional
        Th key to use for sorting; this may return either None or an
        integer key less than the number of elements in the iterable,
        by default None

    Returns
    -------
    List[T]
        A list sorted by the given key
    """
    # Create fallback function if key is undefined
    if key is None:

        def key_(item: T) -> Optional[int]:
            """Return the id of the given object."""
            assert hasattr(item, 'id_')
            val: Optional[int] = item.id_  # type: ignore
            return val

        key = key_
    # Group the input elements into filler and non-filler values
    filler_values = []
    fixed_values = {}
    for item in iterable:
        if key(item) is None:
            filler_values.append(item)
        else:
            fixed_values[key(item)] = item
    # Create a sorted list while filling in the gaps
    output_list = []
    for index, _ in enumerate(iterable):
        if index in fixed_values:
            # Add the item at its appropriate index
            output_list.append(fixed_values[index])
        else:
            # Add a random item to pad the list
            output_list.append(filler_values.pop(0))
    return output_list



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""1D channel generator and processor for BASEmesh.

This module adds support for 1D channels to BASEmesh. These channels
are represented through the :class:`ChannelGeometry` class, which is in
turn made up of :class:`CrossSection` instances.

It also features a number of built-in factories used to simplify the
generation of arbitrary cross sections.

Additionally, 1D geometries may be exported to 2D meshes via the
:meth:`ChannelGeometry.to_mesh()` method.
"""

from . import factories
from .geometry import ChannelGeometry, CrossSection, Vertex
from .io import BaseChainReader, BaseChainWriter

__all__ = [
    'BaseChainReader',
    'BaseChainWriter',
    'ChannelGeometry',
    'CrossSection',
    'factories',
    'io',
    'Vertex'
]

__version__ = '0.1.0a3'



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Python implementation of the quality meshing GUI."""

import logging
import os
from typing import Any, Dict, Iterable, List, Optional
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import (QCheckBox, QComboBox, QGroupBox, QLineEdit,
                             QPlainTextEdit, QPushButton, QRadioButton,
                             QSpinBox, QTextEdit)
from qgis.core import (QgsField, QgsMeshLayer, QgsProject,  # type: ignore
                       QgsVectorLayer)
from qgis.gui import QgisInterface  # type: ignore
from ...core import Lattice
from ...feedback import Feedback
from ...meshing import quality_mesh
from ...stringdefs import write_string_defs_sidecar
from ...triangle import HoleMarker, RegionMarker
from ...types import Line2D, Point2D
from .. import utils
from .base import DialogBase

# NOTE: If the font "Courier New" is unavailable, Qt will pick a different
# monospace font based on the style hint.
_monospace_font = QFont('Courier New')
_monospace_font.setStyleHint(QFont.Monospace)


class QualMeshDialog(DialogBase):
    """GUI implementation of the quality meshing utility."""

    # NOTE: Qt widgets just have a ton of attributes - too bad!
    # pylint: disable=too-many-instance-attributes
    # Qt also uses camelCase for its attributes - too bad!
    # pylint: disable=invalid-name

    ui_file = 'ui_qualmesh.ui'
    help_file = 'help_qualmesh.md'

    def __init__(self, iface: QgisInterface) -> None:
        """Connect signal callbacks and populate widgets."""
        super().__init__(iface)  # Initialize DialogBase

        # NOTE: The following are attribute type hints used to inform supported
        # IDEs of the attribute types defined in the Qt UI file. This code is
        # executed during runtime, but does not do anything as far as the
        # Python interpreter is concerned.
        self.constraintsGroup: QGroupBox
        self.useLines: QCheckBox
        self.linesLayer: QComboBox
        self.useDivCons: QCheckBox
        self.divConsField: QComboBox
        self.usePoints: QCheckBox
        self.pointsLayer: QComboBox
        self.useGlobalMinAngle: QCheckBox
        self.globalMinAngle: QSpinBox
        self.useGlobalMaxArea: QCheckBox
        self.globalMaxArea: QSpinBox
        self.useCustomArgs: QCheckBox
        self.customArgs: QTextEdit
        self.regionsGroup: QGroupBox
        self.regionsLayer: QComboBox
        self.useHole: QCheckBox
        self.holeField: QComboBox
        self.useMatid: QCheckBox
        self.matidField: QComboBox
        self.useMaxArea: QCheckBox
        self.maxAreaField: QComboBox
        self.meshDomainGroup: QGroupBox
        self.keepConvexHull: QRadioButton
        self.shrinkToSegments: QRadioButton
        self.customBoundary: QRadioButton
        self.boundaryLayer: QComboBox
        self.stringDefGroup: QGroupBox
        self.stringDefLayer: QComboBox
        self.stringDefField: QComboBox
        self.includeInMesh: QCheckBox
        self.writeToSidecar: QCheckBox
        self.stringDefBrowseBtn: QPushButton
        self.stringDefPath: QLineEdit
        self.settingsGroup: QGroupBox
        self.logLevel: QComboBox
        self.useSnapping: QCheckBox
        self.snappingTolerance: QSpinBox
        self.outputGroup: QGroupBox
        self.filePath: QLineEdit
        self.fileBrowseBtn: QPushButton
        self.keepTempFiles: QCheckBox
        self.addToMap: QCheckBox
        self.log: QPlainTextEdit
        self.saveLogBtn: QPushButton
        self.clearLogBtn: QPushButton
        self.help: QTextEdit
        self.cancelBtn: QPushButton
        self.runBtn: QPushButton
        self.closeBtn: QPushButton

        self.log.setFont(_monospace_font)
        self.load_help(self.help)

        # Populate combo boxes
        self.add_point_layers(self.regionsLayer, self.pointsLayer)
        self.add_polyline_layers(self.linesLayer, self.stringDefLayer)

        # Hook up callbacks
        self.linesLayer.currentIndexChanged.connect(
            self._update_dividing_constraint)
        self.stringDefLayer.currentIndexChanged.connect(
            self._update_stringdef_field)
        self.saveLogBtn.clicked.connect(
            self.export_log('qualmesh', self.log))
        self.regionsLayer.currentIndexChanged.connect(
            self._update_region_attributes)
        self.stringDefBrowseBtn.clicked.connect(
            self.browse_for_output('txt', self.stringDefPath,
                                   'string-defs'))
        self.fileBrowseBtn.clicked.connect(
            self.browse_for_output('2dm', self.filePath, 'quality-mesh'))

        # Manually trigger some callbacks to pre-populate the combo boxes
        self.linesLayer.currentIndexChanged.emit(0)
        self.regionsLayer.currentIndexChanged.emit(0)
        self.stringDefLayer.currentIndexChanged.emit(0)

    def accept(self) -> None:
        """Evaluate the user's chosen settings and run the tool.

        This function runs when the user confirms their settings, its
        name may not be changed as it is tied directly into the default
        QDialog slot.
        """
        kwargs: Dict[str, Any] = {}

        # Input geometry
        points_data = self.pointsLayer.currentData()
        if self.usePoints.isChecked() and points_data is not None:
            points = list(utils.extract_points(points_data, use_z=False))
        else:
            points = []
        lines_data = self.linesLayer.currentData()
        if self.useLines.isChecked() and lines_data is not None:
            divcon_field: Optional[QgsField] = None
            if self.useDivCons.isChecked():
                divcon_field = self.divConsField.currentData()
            lines = list(utils.extract_lines(lines_data, use_z=False,
                                             divcon_field=divcon_field))
        else:
            lines = []

        # Mesh domain
        if self.shrinkToSegments.isChecked():
            boundary_lines: List[Line2D] = []
        else:
            raise NotImplementedError('Not yet implemented')
        lines.extend(boundary_lines)

        # Region markers
        region_markers: List[RegionMarker] = []
        hole_markers: List[HoleMarker] = []
        if self.regionsGroup.isChecked():
            markers_layer = self.regionsLayer.currentData()
            reg = {}
            if self.useMaxArea.isChecked():
                reg['area_field'] = self.maxAreaField.currentData()
            if self.useMatid.isChecked():
                reg['matid_field'] = self.matidField.currentData()
            if self.useHole.isChecked():
                reg['hole_field'] = self.holeField.currentData()
            region_markers, hole_markers = utils.process_markers(
                markers_layer, **reg)

        # String definitions
        stringdef_layer = self.stringDefLayer.currentData()
        if self.stringDefGroup.isChecked() and stringdef_layer is not None:
            stringdef_field = self.stringDefField.currentData()
            lines.extend(utils.extract_lines(stringdef_layer, use_z=False))
        else:
            stringdef_field = None

        # Precision
        precision = 0.0
        if self.useSnapping.isChecked():
            precision = float(10 ** int(self.snappingTolerance.value()))

        # Parameters
        max_area: float = (self.globalMaxArea.value()
                           if self.useGlobalMaxArea.isChecked() else -1.0)
        min_angle: float = (self.globalMinAngle.value()
                            if self.useGlobalMinAngle.isChecked() else -1.0)
        kwargs['custom_args'] = (self.customArgs.text()
                                 if self.useCustomArgs.isChecked() else '')
        log_level = self.logLevel.currentIndex()
        # NOTE: 0 -> Basic, 1 -> Advanced, 2-4 -> Debug 1-3
        if log_level > 0:
            kwargs['debug_level'] = log_level
            if log_level > 1:
                logging.getLogger('triangle').setLevel(logging.DEBUG)

        # Output
        output_path = self.filePath.text()
        if self.keepTempFiles.isChecked():
            kwargs['move_tempfiles'] = os.path.dirname(output_path)
        else:
            kwargs['move_tempfiles'] = None
        add_file_to_map = self.addToMap.isChecked()

        if min_angle > 0.0:
            kwargs['min_angle'] = min_angle
        if max_area > 0.0:
            kwargs['max_area'] = max_area

        def after(_: Any, error: Optional[BaseException]) -> None:
            """Run after the worker has finished.

            Parameters
            ----------
            _ : Any
                The worker's return value; discarded and unneeded
            error : Optional[BaseException]
                The error that forced the end of execution, by default
                None
            """
            if error is None and add_file_to_map:
                layer_name = os.path.basename(output_path).rsplit('.', 1)[0]
                layer = QgsMeshLayer(output_path, layer_name, 'mdal')
                QgsProject.instance().addMapLayer(layer)
            self._set_lockout(False)
            if error is not None:
                self._push_status((0.0, 'Error', False))
                self.print_error(error)

        self._set_lockout(True)
        if not self.stringDefGroup.isChecked():
            stringdef_layer = None
        self.run_worker(precision, points, lines, output_path, stringdef_layer,
                        stringdef_field, hole_markers, region_markers,
                        self.includeInMesh.isChecked(),
                        self.writeToSidecar.isChecked(),
                        self.stringDefPath.text(), log=self.log,
                        after=after, cancel_btn=self.cancelBtn, **kwargs)

    @staticmethod
    def worker(feedback: Feedback, precision: float, points: Iterable[Point2D],
               lines: Iterable[Line2D], filename: str,
               stringdefs_layer: Optional[QgsVectorLayer],
               stringdefs_field: Optional[QgsField], holes: List[HoleMarker],
               regions: List[RegionMarker], sd_mesh: bool, sd_sidecar: bool,
               sidecar_file: str, **kwargs: Any) -> None:
        """Perform the operation triggered by the dialog.

        This method is static as it will be carried out by a separate
        thread.

        Parameters
        ----------
        feedback : Feedback
            The feedback object use to communicate with the parent
            thread; this will be auto-filled internally
        precision : float
            The precision to use when generating the lattice
        points : Iterable[Tuple[float, float]]
            Input points to triangulate
        lines : Iterable[Tuple[Tuple[float, float],
                         Tuple[float, float]]]
            Input break lines to triangulate
        filename : str
            The output path of the generated mesh
        stringdefs_layer : Optional[QgsVectorLayer]
            The layer containing string definitions
        stringdefs_field : Optional[QgsField]
            The field of the stringdefs layer to use as an identifier
        holes : List[HoleMarker]
            Any hole markers found
        regions : List[RegionMarker]
            Any region markers found
        sd_mesh : bool
            Whether to write string definitions into the mesh itself
        sd_sidecar : bool
            Whether to write string definitions into a sidecare file
        sidecar_file : str
            The name of the sidecar file to write string definitions
            into
        """

        def process_input_data(precision: float, points: Iterable[Point2D],
                               lines: Iterable[Line2D], feedback: Feedback
                               ) -> Lattice:
            """Consolidate the given input geometry into a lattice.

            Parameters
            ----------
            precision : float
                The precision to use when generating the lattice
            points : Iterable[Tuple[float, float]]
                Points to add to the lattice
            lines : Iterable[Tuple[float, float], Tuple[float, float]]
                Lines to add to the lattice
            feedback : Feedback
                The feedback object used to communicate with the parent
                thread

            Returns
            -------
            Lattice
                A lattice containing the processed input data
            """
            # Instantiate lattice. This object will hold and process all of the
            # input geometries.
            lattice = Lattice(precision=precision)

            # Load input points into lattice
            with feedback.busy('Adding geometries to lattice...'):
                print('Reading points...')
                for point in points:
                    lattice.add_node((*point, 0.0), auto_conform=False)
                # Load input line segments into lattice
                print('Reading break lines...')
                for line in lines:
                    nodes = [lattice.add_node((*p, 0.0), auto_conform=False)
                             for p in line]
                    lattice.add_segment(nodes[0], nodes[1], auto_conform=False)

            # Conform input data. This removes duplicates, splits segments, etc.
            print('Conforming input geometries...')
            lattice.conform(feedback=feedback)

            return lattice

        total_steps = '2' if stringdefs_layer is None else '3'

        print('Processing input geometries\n---------------------------')
        feedback.set_prefix(f'Step 1 of {total_steps}: ')
        feedback.set_scaling(0.5)
        lattice = process_input_data(precision, points, lines, feedback)

        print('\n\nRunning Triangle\n----------------\n')
        feedback.set_prefix(f'Step 2 of {total_steps}: ')
        with feedback.busy('Running Triangle'):
            mesh = quality_mesh(*lattice.as_pslg(), holes, regions, **kwargs)

        if stringdefs_layer is not None:
            print('\n\nProcessing string definitions\n'
                  '-----------------------------\n')
            feedback.set_prefix(f'Step 3 of {total_steps}: ')
            feedback.set_scaling(0.5, 0.5)
            feedback.update(0.5, 'Processing string definitions...')
            string_defs = utils.process_string_defs(
                stringdefs_layer, stringdefs_field, mesh, precision,
                feedback=feedback)
            print(f'Found {len(string_defs)} string definitions')
            if sd_mesh:
                print('Writing node strings to mesh file...')
                for name, nodes in string_defs.items():
                    mesh.add_node_string(
                        name, [mesh.get_node_by_id(i) for i in nodes])
            if sd_sidecar:
                print('Writing node strings to sidecar file...')
                write_string_defs_sidecar(string_defs, sidecar_file)

        mesh.save(filename)

        feedback.clear_prefix()
        feedback.clear_scaling()
        feedback.update(1.0, 'Done')

    def _set_lockout(self, is_locked: bool) -> None:
        """Lock out the GUI while running the tool.

        Parameters
        ----------
        is_locked : bool
            Whether to lock or unlock the GUI
        """
        self.constraintsGroup.setEnabled(not is_locked)
        self.regionsGroup.setEnabled(not is_locked)
        self.meshDomainGroup.setEnabled(not is_locked)
        self.stringDefGroup.setEnabled(not is_locked)
        self.settingsGroup.setEnabled(not is_locked)
        self.outputGroup.setEnabled(not is_locked)
        self.runBtn.setEnabled(not is_locked)
        self.closeBtn.setEnabled(not is_locked)
        self.cancelBtn.setEnabled(is_locked)

    def _update_dividing_constraint(self) -> None:
        """Refresh the dividing constraints attribute combo box."""
        breaklines_layer = self.linesLayer.currentData()
        divcon_box = self.divConsField
        divcon_box.clear()
        if breaklines_layer is None:
            return
        for field in breaklines_layer.dataProvider().fields():
            if field.isNumeric() and field.precision() == 0:
                divcon_box.addItem(field.name())
                divcon_box.setItemData(divcon_box.count()-1, field)

    def _update_region_attributes(self) -> None:
        """Refresh the region attribute combo boxes."""
        regions_layer = self.regionsLayer.currentData()
        area_box = self.maxAreaField
        hole_box = self.holeField
        matid_box = self.matidField
        # Clear boxes
        area_box.clear()
        hole_box.clear()
        matid_box.clear()
        if regions_layer is None:
            return
        # Process fields
        for field in regions_layer.dataProvider().fields():
            if not field.isNumeric():
                continue
            area_box.addItem(field.name())
            area_box.setItemData(area_box.count()-1, field)
            if field.precision() == 0:
                hole_box.addItem(field.name())
                hole_box.setItemData(hole_box.count()-1, field)
                matid_box.addItem(field.name())
                matid_box.setItemData(matid_box.count()-1, field)

    def _update_stringdef_field(self) -> None:
        """Refresh the string def name attribute combo box."""
        stringdefs_layer = self.stringDefLayer.currentData()
        sd_box = self.stringDefField
        sd_box.clear()
        if stringdefs_layer is None:
            return
        for field in stringdefs_layer.dataProvider().fields():
            if not field.isNumeric():
                sd_box.addItem(field.name())
                sd_box.setItemData(sd_box.count()-1, field)



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Legacy BASEmesh geometry converter module.

This module defines processing algorithms that convert the geometries
in BASEmesh v1.4 to their respective BASEmesh v2 format.
"""

from typing import Any, Dict, List, Optional
from qgis.core import (QgsFeature, QgsFeatureSink, QgsField,  # type: ignore
                       QgsLineString, QgsPoint, QgsProcessing,
                       QgsProcessingAlgorithm, QgsProcessingContext,
                       QgsProcessingException, QgsProcessingFeedback,
                       QgsProcessingParameterFeatureSink,
                       QgsProcessingParameterField,
                       QgsProcessingParameterNumber,
                       QgsProcessingParameterVectorLayer, QgsVectorLayer,
                       QgsWkbTypes)
from ....algorithms import point_within_range
from ....types import Point3D
from ..utils import parse_docstring


class LegacyLineTo3DLine(QgsProcessingAlgorithm):
    """Converts 2D lines to 3D lines using an auxiliary point layer.

    The elevation of the vertices of each line in the input layer will
    be set to those of the corresponding point in the points layer.

    The snapping tolerance parameter can be used to account for poorly
    aligned points. It will not deduplicate the input lines or perform
    any form of clean-up.

    This can be used to convert input data used for BASEmesh 1.X to the
    format expected by BASEmesh 2.
    """

    # Parameter literals

    INPUT_LINES_LAYER = 'INPUT_LINES_LAYER'
    INPUT_POINTS_LAYER = 'INPUT_POINTS_LAYER'
    INPUT_POINTS_FIELD = 'INPUT_POINTS_FIELD'
    PRECISION = 'PRECISION'
    OUTPUT = 'OUTPUT'

    # Constructor

    @classmethod
    def createInstance(cls) -> 'LegacyLineTo3DLine':
        """Return a new copy of the algorithm."""
        return cls()

    # QGIS properties

    @staticmethod
    def name() -> str:
        """Return the unique algorithm name."""
        return 'convertersconvertlegacylayerline'

    @staticmethod
    def displayName() -> str:
        """Return the user-facing algorithm name."""
        return 'Convert legacy layer (Line)'

    @staticmethod
    def group() -> str:
        """Return the localised name of the algorithm group."""
        return 'Converters'

    @staticmethod
    def groupId() -> str:
        """Return the unique group ID for the algorithm."""
        return 'converters'

    @classmethod
    def shortHelpString(cls) -> str:
        """Return a short help string for the algorithm."""
        return parse_docstring(cls.__doc__)

    def initAlgorithm(self, _: Optional[Dict[str, Any]] = None) -> None:
        """Define the algorithm's inputs and outputs."""
        # Input lines layer
        self.addParameter(QgsProcessingParameterVectorLayer(
            name=self.INPUT_LINES_LAYER,
            description='2D input line layer',
            types=[QgsProcessing.TypeVectorLine]))
        # Input points layer
        self.addParameter(QgsProcessingParameterVectorLayer(
            name=self.INPUT_POINTS_LAYER,
            description='2D input point layer',
            types=[QgsProcessing.TypeVectorPoint]))
        # Input points field
        self.addParameter(QgsProcessingParameterField(
            name=self.INPUT_POINTS_FIELD,
            description='Elevation attribute of the points layer',
            parentLayerParameterName=self.INPUT_POINTS_LAYER,
            type=QgsProcessingParameterField.Numeric))
        # Snapping tolerance
        self.addParameter(QgsProcessingParameterNumber(
            name=self.PRECISION,
            description='Snapping tolerance (exponent)',
            defaultValue=-6,
            minValue=-10,
            maxValue=0))
        # Output
        self.addParameter(QgsProcessingParameterFeatureSink(
            name=self.OUTPUT,
            description='3D output line layer',
            type=QgsProcessing.TypeVectorLine))

    def processAlgorithm(self, parameters: Dict[str, Any],
                         context: QgsProcessingContext,
                         feedback: QgsProcessingFeedback) -> Dict[str, Any]:
        """Execute the algorithm using the given parameters.

        Parameters
        ----------
        parameters : Dict[str, Any]
            The parameters defined for this algorithm
        context : QgsProcessingContext
            The context in which the algorithm is run
        feedback : QgsProcessingFeedback
            The feedback object to use for status reporting

        Returns
        -------
        Dict[str, Any]
            The algorithms output parameters

        Raises
        ------
        QgsProcessingException
            Raised if a line vertex does not match any input point
        """
        lines_layer: QgsVectorLayer = self.parameterAsVectorLayer(
            parameters, self.INPUT_LINES_LAYER, context)
        points_layer: QgsVectorLayer = self.parameterAsVectorLayer(
            parameters, self.INPUT_POINTS_LAYER, context)
        points_field: QgsField = self.parameterAsFields(
            parameters, self.INPUT_POINTS_FIELD, context)[0]
        exponent = int(self.parameterAsDouble(
            parameters, self.PRECISION, context))
        precision = 10 ** exponent
        sink, dest_id = self.parameterAsSink(
            parameters, self.OUTPUT, context, fields=lines_layer.fields(),
            geometryType=QgsWkbTypes.LineStringZ, crs=lines_layer.crs())

        # Read input points
        anchors: List[Point3D] = []
        feat: QgsFeature
        for feat in points_layer.getFeatures():
            # Report status
            if feedback.isCanceled():
                break

            point = feat.geometry().asPoint()
            attr_index = feat.fields().indexFromName(points_field)
            elevation = float(feat.attributes()[attr_index])
            anchors.append((*point, elevation))  # type: ignore

        # Process lines

        feature_count = lines_layer.featureCount()

        old_feat: QgsFeature
        for index, old_feat in enumerate(lines_layer.getFeatures()):
            # Report status
            if feedback.isCanceled():
                break
            feedback.setProgress(int(100*feature_count/(index+1)))

            # Iterate over the points in the (poly-) line
            line_string: List[QgsPoint] = []
            vertex: QgsPoint
            for vertex in old_feat.geometry().vertices():

                # Match point to anchors
                point_2d = vertex.x(), vertex.y()
                found_point: Optional[Point3D] = None
                for anchor in anchors:
                    if point_within_range(point_2d, anchor[:2], precision):
                        found_point = vertex.x(), vertex.y(), anchor[2]
                        break
                if found_point is None:
                    raise QgsProcessingException(
                        f'Unable to match line vertex {point_2d} to an '
                        'elevated point given the precision')

                line_string.append(QgsPoint(*point))

            # Add the 3D line string
            new_feat = QgsFeature()
            new_feat.setGeometry(QgsLineString(line_string))
            new_feat.setAttributes(old_feat.attributes())
            sink.addFeature(new_feat, QgsFeatureSink.FastInsert)

        feedback.setProgress(100)

        return {self.OUTPUT: dest_id}


class LegacyPointTo3DPoint(QgsProcessingAlgorithm):
    """Converts 2D points with elevation attributes to 3D points.

    This algorithm reads each point in the input layer and creates a
    corresponding point in the output.

    The X and Y coordinates are copied from the source point, the Z
    coordinate will be read from the attribute provided.

    This can be used to convert input data used for BASEmesh 1.X to the
    format expected by BASEmesh 2.
    """

    # Parameter literals

    INPUT_LAYER = 'INPUT_LAYER'
    INPUT_FIELD = 'INPUT_FIELD'
    OUTPUT = 'OUTPUT'

    # Constructor

    @classmethod
    def createInstance(cls) -> 'LegacyPointTo3DPoint':
        """Return a new copy of the algorithm."""
        return cls()

    # QGIS properties

    @staticmethod
    def name() -> str:
        """Return the unique algorithm name."""
        return 'convertersconvertlegacylayerpoint'

    @staticmethod
    def displayName() -> str:
        """Return the user-facing algorithm name."""
        return 'Convert legacy layer (Point)'

    @staticmethod
    def group() -> str:
        """Return the localised name of the algorithm group."""
        return 'Converters'

    @staticmethod
    def groupId() -> str:
        """Return the unique group ID for the algorithm."""
        return 'converters'

    @classmethod
    def shortHelpString(cls) -> str:
        """Return a short help string for the algorithm."""
        return parse_docstring(cls.__doc__)

    def initAlgorithm(self, _: Optional[Dict[str, Any]] = None) -> None:
        """Define the algorithm's inputs and outputs."""
        # Input layer
        self.addParameter(QgsProcessingParameterVectorLayer(
            name=self.INPUT_LAYER,
            description='2D input point layer',
            types=[QgsProcessing.TypeVectorPoint]))
        # Input field
        self.addParameter(QgsProcessingParameterField(
            name=self.INPUT_FIELD,
            description='Elevation attribute of the input layer',
            parentLayerParameterName=self.INPUT_LAYER,
            type=QgsProcessingParameterField.Numeric))
        # Output
        self.addParameter(QgsProcessingParameterFeatureSink(
            name=self.OUTPUT,
            description='3D output point layer',
            type=QgsProcessing.TypeVectorPoint))

    def processAlgorithm(self, parameters: Dict[str, Any],
                         context: QgsProcessingContext,
                         feedback: QgsProcessingFeedback) -> Dict[str, Any]:
        """Execute the algorithm using the given parameters."""
        input_layer: QgsVectorLayer = self.parameterAsVectorLayer(
            parameters, self.INPUT_LAYER, context)
        input_field: QgsField = self.parameterAsFields(
            parameters, self.INPUT_FIELD, context)[0]
        sink, dest_id = self.parameterAsSink(
            parameters, self.OUTPUT, context, fields=input_layer.fields(),
            geometryType=QgsWkbTypes.PointZ, crs=input_layer.crs())

        # Retrieve a list of all 3D points in the layer
        feature_count = input_layer.featureCount()

        old_feat: QgsFeature
        for index, old_feat in enumerate(input_layer.getFeatures()):
            # Report status
            if feedback.isCanceled():
                break
            feedback.setProgress(int(100*feature_count/(index+1)))

            # Process old feature
            point = old_feat.geometry().asPoint()
            attr_index = old_feat.fields().indexFromName(input_field)
            elevation = old_feat.attributes()[attr_index]

            # Add new feature
            new_feat = QgsFeature()
            new_feat.setGeometry(QgsPoint(*point, elevation))
            new_feat.setAttributes(old_feat.attributes())
            sink.addFeature(new_feat, QgsFeatureSink.FastInsert)

        feedback.setProgress(100)

        return {self.OUTPUT: dest_id}



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Custom container and iterator definitions for BASEmesh."""

from typing import Any, Generic, Iterable, Iterator, Optional, Set, TypeVar
from .abc import SpatialCollection

T = TypeVar('T')


class SpatialSet(SpatialCollection[T], Generic[T], Set[T]):
    """A set subclass compatible with the SpatialCollection ABC.

    This behaves exactly like a normal set and is used as a temporary
    container until proper space-aware containers are added.

    Please note that this is a temporary solution that will not be
    satisfactory for large data sets.
    It can be considered "deprecated by design" and will be replaced
    with a proper solution very soon.
    """

    def __init__(self, iterable: Optional[Iterable[T]] = None) -> None:
        """Initialise the container.

        Parameters
        ----------
        iterable : Iterable[T], optional
            An iterable to populate the set with, by default None
        """
        self._data: Set[T] = set()
        if iterable is not None:
            self._data.update(iterable)

    def __contains__(self, element: Any) -> bool:
        """Return whether the element is contained.

        Parameters
        ----------
        element : Any
            The element to check for containment

        Returns
        -------
        bool
            Whether the element is contained in the collection
        """
        return element in self._data

    def __iter__(self) -> Iterator[T]:
        """Return an iterator over the container's elements."""
        return iter(self._data)

    def __len__(self) -> int:
        """Return the number of elements in the container."""
        return len(self._data)

    def add(self, value: T) -> None:
        """Add the given value to the collection.

        Parameters
        ----------
        value : T
            The value to add
        """
        self._data.add(value)

    def clear(self) -> None:
        """Remove all items from the container."""
        self._data.clear()

    def discard(self, element: T) -> None:
        """Remove an item from the container.

        If the item cannot be found, do nothing.

        Parameters
        ----------
        element : T
            Discard an element from the collection
        """
        self._data.discard(element)

    def remove(self, element: T) -> None:
        """Remove an element from the container.

        If the element cannot be found, a KeyError will be raised.

        Parameters
        ----------
        element : T
            The element to remove
        """
        self._data.remove(element)

    def pop(self) -> T:
        """Remove and return any item from the container."""
        return self._data.pop()



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Generic geometry classes used for Triangle I/O."""

import contextlib
import itertools
import warnings
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from ..algorithms import (dist_2d, dist_3d, interpolate_line,
                          line_segments_intersection, line_intersection,
                          point_on_line, point_within_range)
from ..errors import TopologyError
from ..feedback import Feedback
from ..log import logger
from ..triangle import PSLGNode, PSLGSegment
from ..types import Line2D, Line3D, Point2D, Point3D


class Node:
    """A point in space.

    A node is a 3D point with optional attributes used to store
    additional information.
    """

    __slots__ = ('attributes', '_attached_segments', 'pos_x', 'pos_y', 'pos_z')

    def __init__(self, pos: Union[Point2D, Point3D], **kwargs: Any) -> None:
        """Create a new node from the given point.

        Any keyword arguments passed will be treated as attributes and
        stored alongside the node.

        Parameters
        ----------
        pos : Union[Tuple[float, float], Tuple[float, float, float]]
            The position of the node
        """
        self.pos_x = pos[0]
        self.pos_y = pos[1]
        try:
            self.pos_z = pos[2]  # type: ignore
        except IndexError:
            self.pos_z = 0.0
        self.attributes: Dict[str, Any] = OrderedDict(
            sorted(kwargs.items(), key=lambda x: x[0].lower()))
        self._attached_segments: Set['Segment'] = set()

    def __del__(self) -> None:
        """Delete any segments along with their defining node.

        Note that any exceptions raised herein will not interrupt the
        deletion process.
        """
        # Delete any attached geometries along with their defining node
        while self._attached_segments:
            with contextlib.suppress(KeyError):
                _ = self._attached_segments.pop()

    @property
    def attached_segments(self) -> Set['Segment']:
        """Return the set of attached segments."""
        return self._attached_segments

    @property
    def elevation(self) -> float:
        """Return the elevation of the node."""
        return self.pos_z

    def attach_segment(self, segment: 'Segment') -> None:
        """Register the given segment as being attached to the node.

        This is usually done as part of the Segment class's
        initialiser and is used to ensure the segment will be deleted
        with its defining node.

        Parameters
        ----------
        segment : Segment
            The segment instance to attach to the node
        """
        self._attached_segments.add(segment)

    def detach_segment(self, segment: 'Segment') -> None:
        """Unregister the given segment from the node.

        This is usually done to ensure the Segment object can be
        garbage-collected when it is no longer needed, since keeping it
        referenced in a node would prevent its destruction.

        Parameters
        ----------
        segment : Segment
            The segment instance to detach from the node
        """
        self._attached_segments.discard(segment)

    # pylint: disable=invalid-name
    def move(self, x: Optional[float] = None, y: Optional[float] = None,
             z: Optional[float] = None) -> None:
        """Move the node to a new location.

        Any coordinate whos argument is left at None will keep their
        current value.
        Note that this also mvoes any geometries defined through this
        node.

        Parameters
        ----------
        x : float, optional
            New X coordinate of the node, by default None
        y : float, optional
            New Y coordinate of the node, by default None
        z : float, optional
            New Z coordinate of the node, by default None
        """
        if x is not None:
            self.pos_x = x
        if y is not None:
            self.pos_y = y
        if z is not None:
            self.pos_z = z

    def as_tuple(self) -> Tuple[float, float, float]:
        """Return the coordinates of the node as a tuple."""
        return self.pos_x, self.pos_y, self.pos_z

    def as_tuple_2d(self) -> Tuple[float, float]:
        """Return the X and Y coordinates of the node as a tuple."""
        return self.pos_x, self.pos_y


class Segment:
    """A 3D line segment defined through two nodes."""

    __slots__ = ('_node_1', '_node_2')

    def __init__(self, node_1: Node, node_2: Node) -> None:
        """Initialise the line segment from two nodes.

        Note that the nodes will be sorted, the order of the arguments
        is therefore inconsequential.

        Parameters
        ----------
        node_1 : Node
            First input node
        node_2 : Node
            Second input node

        Raises
        ------
        TopologyError
            Raised if the two nodes are coincident or identical
        TopologyError
            Raised if the X and Y coordinates of the input nodes are
            identical, even if their elevation is different
        """
        if node_1 == node_2:
            raise TopologyError(
                'Two separate nodes required to define a segment')
        self._node_1 = node_1
        self._node_2 = node_2
        # Re-order nodes if needed
        if self._nodes_sorted() != (self._node_1, self._node_2):
            self._node_2 = node_1
            self._node_1 = node_2

        self._node_1.attach_segment(self)
        self._node_2.attach_segment(self)
        if self.length_2d == 0:
            raise TopologyError(
                'The 2D projection of the segment has a length of zero')

    def __eq__(self, value: Any) -> bool:
        """Segment equality check.

        This ensures that two segments connecting the same nodes will
        be considered identical regardless of node order.

        Parameters
        ----------
        value : Any
            The value to compare the segment against

        Returns
        -------
        bool
            Whether the two objects are identical
        """
        if not isinstance(value, self.__class__):
            return False
        return self.nodes == value.nodes

    def __hash__(self) -> int:
        """Return the hash for the given segment.

        This uses a coordinate-sorted list for the nodes to ensure two
        segments of opposing direction will hash the same.
        """
        return hash((self._node_1, self._node_2))

    @property
    def length(self) -> float:
        """Return the 3D length of the segment."""
        return dist_3d(*self.as_line())

    @property
    def length_2d(self) -> float:
        """Return the 2D length of the segment."""
        return dist_2d(*self.as_line_2d())

    @property
    def nodes(self) -> Set[Node]:
        """Return the nodes defining this segment.

        The returned set will always have exactly two elements.
        """
        return set((self._node_1, self._node_2))

    def as_line(self) -> Line3D:
        """Return the segment as a tuple of 3D points.

        The returned points are coordinate-sorted.
        """
        n_1 = self._node_1
        n_2 = self._node_2
        return ((n_1.pos_x, n_1.pos_y, n_1.pos_z),
                (n_2.pos_x, n_2.pos_y, n_2.pos_z))

    def as_line_2d(self) -> Line2D:
        """Return the segment as a tuple of 2D points.

        The returned points are coordinate-sorted.
        """
        n_1 = self._node_1
        n_2 = self._node_2
        return (n_1.pos_x, n_1.pos_y), (n_2.pos_x, n_2.pos_y)

    def _nodes_sorted(self) -> Tuple[Node, Node]:
        """Sort the segment's nodes by coordinates.

        This is done to ensure that two segments defined using opposite
        node order end up as the same object.

        Returns
        -------
        Tuple[Node, Node]
            The segment's nodes sorted by its coordinates
        """
        # NOTE: This orients the line so it faces right and up
        pt_a = self._node_1.as_tuple_2d()
        pt_b = self._node_2.as_tuple_2d()
        # Sort using X
        if pt_a[0] < pt_b[0]:
            return self._node_1, self._node_2
        if pt_a[0] > pt_b[0]:
            return self._node_2, self._node_1
        # X coordinate identical -> sort using Y
        if pt_a[1] < pt_b[1]:
            return self._node_1, self._node_2
        return self._node_2, self._node_1

    def split(self, node: Node) -> 'Segment':
        """Split the segment at the given node.

        The current segment is shortened to end at the given node, and
        a new segment starting at the inserted node and ending at the
        current node's original end point is created and returned.

        Returns
        -------
        Segment
            The new segment that was created, leading from the
            inserted node to this Segment's original end point

        Raises
        ------
        TopologyError
            Raised if the given node is coincident with one of the
            segment's defining nodes
        """
        if node in (self._node_1, self._node_2):
            raise TopologyError(
                'Unable to split segment on one of its defining nodes')
        # Store the original end node
        end_node = self._node_2
        self._node_2.detach_segment(self)
        # Move this segment to end at the inserted node
        self._node_2 = node
        self._node_2.attach_segment(self)
        # Create a new segment (this also attaches the segment to the nodes)
        return Segment(node, end_node)


class Lattice:
    """An unstructured 3D collection of nodes and segments."""

    def __init__(self, precision: float = 1e-6) -> None:
        """Instantiate the lattice.

        The precision is used for any non-exact checks for all geometry
        added to the lattice container.

        Parameters
        ----------
        precision : float, optional
            The snapping tolerance for any geometries added, by default
            1e-6
        """
        self._nodes: List[Node] = []
        self.precision = precision
        self._segments: List[Segment] = []

    @property
    def nodes(self) -> List[Node]:
        """Return the list of nodes in the lattice."""
        return self._nodes

    @property
    def segments(self) -> List[Segment]:
        """Return the list of segmetns in the lattice."""
        return self._segments

    def add_node(self, point: Point3D, warn_duplicate: bool = False,
                 auto_conform: bool = True) -> Node:
        """Add a new node to the lattice.

        If a node already exists at the position of the given point, no
        node is created. If warn_duplicate is enabled, a warning will
        be broadcast.

        Parameters
        ----------
        point : Tuple[float, float, float]
            The location of the node to add
        warn_duplicate : bool, optional
            Whether to broadcast a warning if a node already exists at
            this 2D location, by default False
        auto_conform : bool, optional
            Whether to conform the lattice after adding the node, by
            default True

        Returns
        -------
        Node
            The created node, or the existing node if one already
            existed at this location
        """
        for node in self.nodes:
            # If an existing node is close enough
            if point_within_range(node.as_tuple_2d(), point[:2],
                                  max_dist=self.precision):
                if warn_duplicate:
                    warnings.warn(f'Duplicate node at {node.as_tuple_2d()}')
                logger.debug('Found existing node at %s', node.as_tuple_2d())
                return node

        logger.debug('Creating new node at %s', point[:2])
        node = Node(point)
        self.nodes.append(node)
        if auto_conform:
            self.conform()

        return node

    def add_segment(self, node_1: Node, node_2: Node,
                    warn_duplicate: bool = False,
                    auto_conform: bool = True,) -> None:
        """Add a new segment to the lattice.

        If another segment already exists between the given nodes, no
        segment is created. If warn_duplicate is enabled, a warning
        will be broadcast.

        Parameters
        ----------
        node_1 : Node
            The first node defining the segment
        node_2 : Node
            The second node defining the segoment
        warn_duplicate : bool, optional
            Whether to broadcast a warning if a segment already exists
            between these points, by default False
        auto_conform : bool, optional
            Whether to conform the lattice after adding the segment,
            by default True
        """
        # Only create the segment if it is not an exact duplicate
        for segment in self.segments:
            if segment.nodes == set((node_1, node_2)):
                line = tuple(n.as_tuple_2d() for n in segment.nodes)
                if warn_duplicate:
                    warnings.warn(f'Duplicate segment at {line}')
                logger.debug('A segment between %s and %s already exists',
                             line[0], line[1])
                return
        # Only create the segment if it not degenerate
        if node_1 == node_2:
            pos = node_1.as_tuple_2d()
            if warn_duplicate:
                warnings.warn(f'Degenerate segment at {pos}')
            logger.debug('Zero-length segment encountered at %s', pos)
            return

        new_segment = Segment(node_1, node_2)
        logger.debug('Creating new segment at %s',
                     tuple(n.as_tuple_2d() for n in new_segment.nodes))
        if new_segment not in self.segments:
            self.segments.append(new_segment)
        if auto_conform:
            self.conform()

    def as_pslg(self) -> Tuple[List[PSLGNode], List[PSLGSegment]]:
        """Return the contents of the lattice in Triangle format.

        The user is responsible for ensuring their geometry is valid,
        either by taking care when adding geometry or by calling
        Lattice.conform() before calling this function.

        Since Triangle does not support 3D geometries, the node
        elevation will be stored in the "pos_z" attribute of the
        PSLGNode class.

        Returns
        -------
        Tuple[List[PSLGNode], List[PSLGSegment]]
            The nodes and segments representing this lattice
        """
        triangle_nodes: List[PSLGNode] = [
            PSLGNode(*n.as_tuple_2d(), pos_z=n.elevation) for n in self.nodes]
        triangle_segments: List[PSLGSegment] = [
            PSLGSegment.from_points(*l.as_line(), nodes=triangle_nodes)
            for l in self.segments]
        return triangle_nodes, triangle_segments

    def conform(self, *, feedback: Optional[Feedback] = None) -> None:
        """Process the lattice's geometries to make them compliant.

        This involves node deduplication, insertion of nodes at segment
        intersections and splitting segments via nodes.

        Note that this can be a fairly expensive operation for large
        geometries.

        Parameters
        ----------
        feedback : Feedback, optional
            A helper object used to retrieve status information about
            the conformation process, by default None
        """
        # Step 1: Deduplicate nodes
        if feedback is not None:
            feedback.update(msg='Deduplicating nodes...')
            feedback.set_scaling(0.2)
        removed = self.deduplicate_nodes(feedback=feedback)
        print(f'{removed} nodes removed')

        # Step 2: Find intersections
        if feedback is not None:
            feedback.update(msg='Finding intersections...')
            feedback.set_scaling(0.6, 0.2)
        added = self.add_intersection_nodes(feedback=feedback)
        print(f'{added} nodes added')

        # Step 3: Split segments
        if feedback is not None:
            feedback.update(msg='Splitting segments...')
            feedback.set_scaling(0.2, 0.8)
        splits = self.split_segments(feedback=feedback)
        print(f'{splits} segments split')

    def deduplicate_nodes(self, *, feedback: Optional[Feedback] = None) -> int:
        """Remove and duplicate nodes.

        Parameters
        ----------
        feedback : Feedback, optional
            The Feedback object to use for status updates, by default
            None

        Returns
        -------
        int
            The number of nodes removed.
        """
        nodes_to_remove: List[Node] = []
        # Get a list of all unique, unordered pairs of nodes
        combinations: List[Tuple[Node, Node]] = list(
            itertools.combinations(self.nodes, 2))  # type: ignore
        if feedback is not None:
            total = len(combinations)
            interval = int(total / 100) if total > 100 else 1

        element: Tuple[Node, Node]
        for index, element in enumerate(combinations):
            node_a, node_b = element
            pt_a = node_a.as_tuple_2d()
            pt_b = node_b.as_tuple_2d()

            if point_within_range(pt_a, pt_b, max_dist=self.precision):
                logger.info('Removing duplicate node at %s', pt_b)

                # Relink node segments
                # NOTE: list cast used to allow mutation of the underlying set
                for segment in list(node_b.attached_segments):
                    logger.debug('Relinking segment %s', segment.as_line_2d())
                    self.segments.remove(segment)
                    node_b.detach_segment(segment)

                    # Create a copy of the node
                    nodes_list = list(segment.nodes)
                    nodes_list.remove(node_b)
                    other_node = nodes_list[0]
                    segment = Segment(node_a, other_node)
                    self.segments.append(segment)
                    node_a.attach_segment(segment)

                # Mark node segment for deletion
                nodes_to_remove.append(node_b)

            if feedback is not None and index % interval == 0:
                feedback.update((index+1)/total)

        if nodes_to_remove:
            logger.info('Deleting %d duplicate nodes', len(nodes_to_remove))
            _ = (self.nodes.remove(n) for n in nodes_to_remove)  # type: ignore

        # Return the number of deleted nodes
        return len(nodes_to_remove)

    def add_intersection_nodes(self, *, feedback: Optional[Feedback] = None
                               ) -> int:
        """Add a node at each segment intersection point.

        Call split_segments() afterwards to get a non-intersecting
        lattice.

        Parameters
        ----------
        feedback : Feedback, optional
            The Feedback object to use for status updates, by default
            None, by default None

        Returns
        -------
        int
            The number of inserted nodes

        Raises
        ------
        TopologyError
            Raised if two collinear segments are encountered
        """
        combinations = list(itertools.combinations(self.segments, 2))

        if feedback is not None:
            total = len(combinations)
            interval = int(total / 100) if total > 100 else 1

        added_nodes = 0

        for index, tuple_ in enumerate(combinations):
            segment_a, segment_b = tuple_
            line_a = segment_a.as_line_2d()
            line_b = segment_b.as_line_2d()

            if feedback is not None and index % interval == 0:
                feedback.update((index+1)/total)

            if not line_segments_intersection(line_a, line_b,
                                              allow_collinear=False):
                continue

            # Check for coincident endpoints
            if line_a[0] in line_b or line_a[1] in line_b:
                logger.debug('Segments %s and %s share an endpoint (exact), '
                             'not a valid intersection', line_a, line_b)
                continue

            logger.debug('Segments %s and %s intersect, calculating '
                         'intersection point', line_a, line_b)
            try:
                intersection = line_intersection(line_a, line_b)
            except ValueError as err:
                raise TopologyError(f'Collinear line segments found: {line_a} '
                                    f'and {line_b}') from err
            # Ignore endpoint intersections
            is_valid = True
            logger.debug('Checking intersection distance from endpoints')
            for endpoint in (*line_a, *line_b):
                if dist_2d(intersection, endpoint) < self.precision:
                    logger.debug('Segments touching in %s, not a valid '
                                 'intersection', intersection)
                    is_valid = False
                    break
            if not is_valid:
                continue
            logger.info('Segment intersection at %s', intersection)

            # Create an intersection node
            height_a = interpolate_line(intersection, segment_a.as_line())
            height_b = interpolate_line(intersection, segment_b.as_line())
            # Check height difference
            if abs(height_a-height_b) > self.precision:
                warnings.warn(
                    f'Conflicting elevation data for {intersection}: using '
                    f'original height of {height_a} and ignoring other (was '
                    f'{height_b}).\nNote that this causes the line segments '
                    f'{(line_b[0], intersection, line_b[1])} to lose '
                    f'their collinearity.')
            self.add_node((*intersection, height_a), auto_conform=False)
            added_nodes += 1

        return added_nodes

    def split_segments(self, *, feedback: Optional[Feedback] = None) -> int:
        """Split the lattice's segments at the nodes.

        This ensures no node lies in the middle of a segment, and
        segment will lie between two nodes. If a segment is split by
        more than one node, it will be split more than once.

        Parameters
        ----------
        feedback : Feedback, optional
            The Feedback object to use for status updates, by default
            None, by default None, by default None

        Returns
        -------
        int
            The number of splits performed
        """
        if feedback is not None:
            total = len(self.nodes)

        splits_performed = 0

        for index, node in enumerate(self.nodes):
            node_pt = node.as_tuple_2d()
            segment_split = True
            while segment_split:
                segment_split = False

                for segment in self.segments:
                    line = segment.as_line_2d()
                    if (point_on_line(line, node_pt, precision=self.precision)
                            and not node in segment.nodes):
                        new_segment = segment.split(node)
                        self.segments.append(new_segment)
                        segment_split = True
                        splits_performed += 1
                        break

            if feedback is not None:
                feedback.update((index+1)/total)

        return splits_performed



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract base classes for BASEmesh."""

# pylint: disable=too-few-public-methods

from abc import ABCMeta, abstractmethod
from typing import Iterator, Collection, Generic, TypeVar
from .types import Point2D

T = TypeVar('T')


class ElevationSource(metaclass=ABCMeta):
    """ABC for elevation information provider classes.

    Classes supporting this ABC can be used as elevation sources for
    mesh interpolation.
    """

    @abstractmethod
    def elevation_at(self, point: Point2D) -> float:
        """Return the elevation at the given point.

        If the point lies outside the intended area of this elevation
        source, a ValueError should be raised instead. The interpolator
        will then fall through to the next elevation source and attempt
        to use its elevation data.

        Parameters
        ----------
        point : Tuple[float, float]
            The point to interpolate

        Returns
        -------
        float
            The interpolated elevation

        Raises
        ------
        NotImplementedError
            Always raised as this is an abstract method
        """
        raise NotImplementedError


class SpatialCollection(Generic[T], Collection[T]):
    """Represents a spatial container type.

    This ABC adds the iter_spatial() method, which allows optimisation
    of the iteration order using a seed point.

    iter_spatial() will fall back to the default iterator if not
    overridden.
    """

    def iter_spatial(self, point: Point2D) -> Iterator[T]:
        """Iterate over the contents, starting near the seed point.

        Similar to __iter__, but with a predefined starting point
        used for iterator creation. The iterator will return elements
        near the given point before moving outwards.

        Note that there is no guarantee that the order will actually
        hit nearby elements first.

        Parameters
        ----------
        point : Tuple[float, float]
            The seed point to start iterating at

        Returns
        -------
        Iterator[T]
            Iterator over the contents
        """
        _ = point  # Value intentionally discarded
        return iter(self)


class SupportsSpatial(metaclass=ABCMeta):
    """ABC for objects that can be used in space-aware containers.

    This ABC defines the spatial_marker property, which returns a
    tuple of two floats representing the given object.

    Attributes
    ----------
    spatial_marker : Tuple[float, float]
        A 2D point representing the object, used for positioning in
        memory
    """

    @property
    @abstractmethod
    def spatial_marker(self) -> Point2D:
        """Return a 2D point representing this object.

        This point will be used by space-aware containers to optimise
        memory layout.
        """
        raise NotImplementedError



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A Python wrapper for Jonathan Shewchuk's "Triangle".

This module provides a Python API for the Triangle command line
interface, as well as Python representations of its input data types.
"""

from . import triangle_io
from .markers import HoleMarker, RegionMarker
from .pslg import PSLGElement, PSLGNode, PSLGSegment
from .wrapper import Triangle

__all__ = [
    'HoleMarker',
    'PSLGElement',
    'PSLGNode',
    'PSLGSegment',
    'RegionMarker',
    'Triangle',
    'triangle_io'
]

__version__ = '0.1.0'
